# -*- coding: utf-8 -*-
"""
Calculates the spectrum, based on a set of stellar parameters
And also determines the best fit parameters
"""

import json
import logging
import warnings
from numbers import Number, Real
from os.path import splitext

import numpy as np
from scipy.constants import speed_of_light
from scipy.optimize import OptimizeWarning, least_squares
from scipy.optimize._numdiff import approx_derivative
from scipy.stats import norm
from tqdm import tqdm

from . import __file_ending__
from .abund import Abund
from .atmosphere.atmosphere import AtmosphereError
from .atmosphere.krzfile import KrzFile
from .atmosphere.savfile import SavFile
from .large_file_storage import setup_lfs
from .nlte import DirectAccessFile
from .sme import MASK_VALUES
from .synthesize import Synthesizer
from .util import print_to_log, show_progress_bars

logger = logging.getLogger(__name__)

clight = speed_of_light * 1e-3  # km/s
warnings.filterwarnings("ignore", category=OptimizeWarning)


class VariableNumber(np.lib.mixins.NDArrayOperatorsMixin):
    """
    This is essentially a pointer to the value, but will interact with operators
    as though it was the value. This is useful to pass one f_scale parameter
    to the least squares fit, that can then be modified during the iterations.
    """

    def __init__(self, value):
        self.value = value

    _HANDLED_TYPES = (np.ndarray, Number)

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        out = kwargs.get("out", ())
        for x in inputs + out:
            # Only support operations with instances of _HANDLED_TYPES.
            # Use ArrayLike instead of type(self) for isinstance to
            # allow subclasses that don't override __array_ufunc__ to
            # handle ArrayLike objects.
            if not isinstance(x, self._HANDLED_TYPES + (VariableNumber,)):
                return NotImplemented

        # Defer to the implementation of the ufunc on unwrapped values.
        inputs = tuple(x.value if isinstance(x, VariableNumber) else x for x in inputs)
        if out:
            kwargs["out"] = tuple(
                x.value if isinstance(x, VariableNumber) else x for x in out
            )
        result = getattr(ufunc, method)(*inputs, **kwargs)

        if type(result) is tuple:
            # multiple return values
            return result
        elif method == "at":
            # no return value
            return None
        else:
            # one return value
            return result


class SME_Solver:
    def __init__(self, filename=None, restore=False):
        self.config, self.lfs_atmo, self.lfs_nlte = setup_lfs()
        self.synthesizer = Synthesizer(
            config=self.config,
            lfs_atmo=self.lfs_atmo,
            lfs_nlte=self.lfs_nlte,
        )

        # Various parameters to keep track of during solving
        self.filename = filename
        self.iteration = 0
        self.parameter_names = []
        self.update_linelist = False
        self._latest_residual = None
        self._latest_jacobian = None
        self.restore = restore
        self.f_scale = 0.2
        # For displaying the progressbars
        self.progressbar = None
        self.progressbar_jacobian = None

    @property
    def nparam(self):
        return len(self.parameter_names)

    def restore_func(self, sme):
        fname = self.filename.rsplit(".", 1)[0]
        fname = f"{fname}_iter.json"
        try:
            with open(fname) as f:
                data = json.load(f)
            # The keys are string, but we want the max in int, so we need to convert back and forth
            iteration = str(max(int(i) for i in data.keys()))
            for fp in self.parameter_names:
                sme[fp] = data[iteration].get(fp, sme[fp])
            logger.warning(f"Restoring existing backup data from {fname}")
        except:
            pass
        return sme

    def backup(self, sme):
        fname = self.filename.rsplit(".", 1)[0]
        fname = f"{fname}_iter.json"
        try:
            with open(fname) as f:
                data = json.load(f)
        except:
            data = {}
        data[self.iteration] = {fp: sme[fp] for fp in self.parameter_names}
        try:
            with open(fname, "w") as f:
                json.dump(data, f)
        except:
            pass

    def _residuals(
        self, param, sme, spec, uncs, mask, segments="all", isJacobian=False, **_
    ):
        """
        Calculates the synthetic spectrum with sme_func and
        returns the residuals between observation and synthetic spectrum

        residual = (obs - synth) / uncs

        Parameters
        ----------
        param : list(float) of size (n,)
            parameter values to use for synthetic spectrum, order is the same as names
        names : list(str) of size (n,)
            names of the parameters to set, as defined by SME_Struct
        sme : SME_Struct
            sme structure holding all relevant information for the synthetic spectrum generation
        spec : array(float) of size (m,)
            observed spectrum
        uncs : array(float) of size (m,)
            uncertainties of the observed spectrum
        mask : array(bool) of size (k,)
            mask to apply to the synthetic spectrum to select the same points as spec
            The size of the synthetic spectrum is given by sme.wave
            then mask must have the same size, with m True values
        isJacobian : bool, optional
            Flag to use when within the calculation of the Jacobian (default: False)
        fname : str, optional
            filename of the intermediary product (default: "sme.npy")
        fig : Figure, optional
            plotting interface, fig.add(x, y, title) will be called each non jacobian iteration

        Returns
        -------
        resid : array(float) of size (m,)
            residuals of the synthetic spectrum
        """
        update = not isJacobian
        save = not isJacobian and self.filename is not None
        reuse_wavelength_grid = isJacobian
        radial_velocity_mode = "robust" if not isJacobian else "fast"

        # change parameters
        for name, value in zip(self.parameter_names, param):
            sme[name] = value
        # run spectral synthesis
        try:
            result = self.synthesizer.synthesize_spectrum(
                sme,
                updateStructure=update,
                reuse_wavelength_grid=reuse_wavelength_grid,
                segments=segments,
                passLineList=False,
                updateLineList=self.update_linelist,
                radial_velocity_mode=radial_velocity_mode,
            )
        except AtmosphereError as ae:
            # Something went wrong (left the grid? Don't go there)
            # If returned value is not finite, the fit algorithm will not go there
            logger.debug(ae)
            return np.full(spec.size, np.inf)

        segments = Synthesizer.check_segments(sme, segments)

        # Get the correct results for the comparison
        synth = sme.synth if update else result[1]
        synth = synth[segments]
        synth = synth[mask] if mask is not None else synth

        if sme.telluric is not None:
            tell = sme.telluric[segments]
            tell = tell[mask] if mask is not None else tell
            synth = synth * tell

        resid = (synth - spec) / (uncs)
        resid = resid.ravel()
        resid = np.nan_to_num(resid, copy=False)

        # Update progress bars
        if isJacobian:
            self.progressbar_jacobian.update(1)
        else:
            self.progressbar.total += 1
            self.progressbar.update(1)

        if not isJacobian:
            # Update f_scale
            # self.f_scale.value = np.percentile(np.abs(resid), 95)
            # print(self.f_scale.value)
            # Save result for jacobian
            self._latest_residual = resid
            self.iteration += 1
        logger.debug("%s", {n: f"{v:.3f}" for n, v in zip(self.parameter_names, param)})

        # Also save intermediary results, because we can
        if save:
            self.backup(sme)

        return resid

    def _jacobian(
        self,
        param,
        *args,
        bounds=None,
        segments="all",
        step_sizes=None,
        method="2-point",
        **_,
    ):
        """
        Approximate the jacobian numerically
        The calculation is the same as "2-point"
        but we can tell residuals that we are within a jacobian.

        Note that when we reuse the wavelength grid, the results are
        slightly different for reasons(?). Therefore the step size should
        be larger than those differences, which is why we specify the
        step size for each parameter.
        """
        self.progressbar_jacobian.reset()

        # Here we replace the scipy version of approx_derivative with our own
        # The only difference being that we use Multiprocessing for the jacobian
        g = approx_derivative(
            self._residuals,
            param,
            method=method,
            # This feels pretty bad, passing the latest synthetic spectrum
            # by reference as a parameter of the residuals function object
            f0=self._latest_residual,
            abs_step=step_sizes,
            bounds=bounds,
            args=args,
            kwargs={"isJacobian": True, "segments": segments},
        )

        if not np.all(np.isfinite(g)):
            g[~np.isfinite(g)] = 0
            logger.warning(
                "Some derivatives are non-finite, setting them to zero. "
                "Final uncertainties will be inaccurate. "
                "You might be running into the boundary of the grid"
            )
        self._latest_jacobian = np.copy(g)
        return g

    def get_bounds(self, sme):
        """
        Create Bounds based on atmosphere grid and general rules

        Note that bounds define by definition a cube in the parameter space,
        but the grid might not be a cube. I.e. Not all combinations of teff, logg, monh are valid
        This method will choose the outerbounds of that space as the boundary, which means that
        we can still run into problems when interpolating the atmospheres

        Parameters
        ----------
        param_names : array(str)
            names of the parameters to vary
        sme : SME_Structure
            sme structure to get bounds for

        Raises
        ------
        IOError
            If the atmosphere file can't be read, allowed types
            are IDL savefiles (.sav), and .krz files

        Returns
        -------
        bounds : dict
            Bounds for the given parameters
        """

        bounds = {}

        # Create bounds based on atmosphere grid
        if (
            "teff" in self.parameter_names
            or "logg" in self.parameter_names
            or "monh" in self.parameter_names
        ):
            if sme.atmo.method == "grid":
                atmo_source = sme.atmo.source
                _, ext = splitext(atmo_source)
                atmo_file = self.lfs_atmo.get(atmo_source)

                if ext == ".sav":
                    atmo_grid = SavFile(
                        atmo_file, source=atmo_source, lfs=self.lfs_atmo
                    )

                    teff = np.unique(atmo_grid.teff)
                    teff = np.min(teff), np.max(teff)
                    bounds["teff"] = teff

                    logg = np.unique(atmo_grid.logg)
                    logg = np.min(logg), np.max(logg)
                    bounds["logg"] = logg

                    monh = np.unique(atmo_grid.monh)
                    monh = np.min(monh), np.max(monh)
                    bounds["monh"] = monh
                elif ext == ".krz":
                    # krz atmospheres are fixed to one parameter set
                    # allow just "small" area around that
                    atmo = KrzFile(atmo_file, source=atmo_source)
                    bounds["teff"] = atmo.teff - 500, atmo.teff + 500
                    bounds["logg"] = atmo.logg - 1, atmo.logg + 1
                    bounds["monh"] = atmo.monh - 1, atmo.monh + 1
                else:
                    raise IOError(f"File extension {ext} not recognized")
            if sme.atmo.method == "embedded":
                atmo = sme.atmo
                bounds["teff"] = atmo.teff - 500, atmo.teff + 500
                bounds["logg"] = atmo.logg - 1, atmo.logg + 1
                bounds["monh"] = atmo.monh - 1, atmo.monh + 1
        # Add generic bounds
        bounds.update(
            {
                "vmic": [0, clight],
                "vmac": [0, clight],
                "vsini": [0, clight],
                "ipres": [1, 500_000],
            }
        )
        # bounds.update({"abund %s" % el: [-10, 11] for el in abund_elem})

        result = np.array([[-np.inf, np.inf]] * self.nparam)
        solar = Abund.solar()
        for i, name in enumerate(self.parameter_names):
            if name[:5].lower() == "abund":
                element = name[5:].strip().capitalize()
                if element in sme.nlte.elements:
                    fname = sme.nlte.grids[element]
                    fname = self.lfs_nlte.get(fname)
                    grid = DirectAccessFile(fname)
                    available = grid["abund"]
                    xmin, xmax = available.min(), available.max()
                    xmin += solar[element]
                    xmax += solar[element]
                    if xmin == xmax:
                        xmin -= 1
                        xmax += 1
                    result[i] = [xmin, xmax]
                else:
                    result[i] = [-10, 11]
            elif name[:8].lower() == "linelist":
                if "excit" in name.lower():
                    value = sme.linelist[int(name.split()[1])]["excit"][0]
                    result[i] = [value - 0.005, value + 0.005]
                else:
                    result[i] = [-np.inf, np.inf]
            else:
                result[i] = bounds[name]

        result = result.T

        if len(result) > 0:
            return result
        else:
            return [-np.inf, np.inf]

    def get_scale(self):
        """
        Returns scales for each parameter so that values are on order ~1

        Parameters
        ----------
        param_names : list(str)
            names of the parameters

        Returns
        -------
        scales : list(float)
            scales of the parameters in the same order as input array
        """

        # The only parameter we want to scale right now is temperature,
        # as it is orders of magnitude larger than all others
        scales = {"teff": 1000}
        scales = [
            scales[name] if name in scales.keys() else 1
            for name in self.parameter_names
        ]
        return scales

    def get_step_sizes(self, parameter_names):
        steps = {
            "teff": 10,
            "logg": 0.01,
            "monh": 0.01,
            "vmic": 0.01,
            "vmac": 0.1,
            "vsini": 0.05,
            "vrad": 0.05,
            "gam6": 0.02,
            "ipres": 1000,
        }
        step_sizes = []
        for param in parameter_names:
            if param in steps.keys():
                step_sizes += [steps[param]]
            elif param.startswith("Abund"):
                step_sizes += [0.01]
            elif param.startswith("linelist"):
                if "excit" in param:
                    step_sizes += [0.001]
                else:
                    step_sizes += [0.2]
            else:
                step_sizes += [0.001]
        step_sizes = np.asarray(step_sizes)
        return step_sizes

    def get_default_values(self, sme):
        """Default parameter values for each name"""
        d = {"teff": 5778, "logg": 4.4, "monh": 0, "vmac": 1, "vmic": 1}
        d.update({f"{el} abund": v for el, v in Abund.solar()().items()})

        def default(name):
            logger.info("No value for %s set, using default value %s", name, d[name])
            return d[name]

        values = [
            sme[s] if sme[s] is not None else default(s) for s in self.parameter_names
        ]
        return np.array(values)

    @staticmethod
    def estimate_uncertainties(resid, deriv):
        """
        Estimate the uncertainties by fitting the cumulative distribution of
        derivative / uncertainties vs. residual / derivative
        with the generalized normal distribution and use the 68% percentile
        as the 1 sigma approximation for a normally distributed variable

        Parameters
        ----------
        unc : array of shape (n,)
            uncertainties
        resid : array of shape (n,)
            residuals of the least squares fit
        deriv : array of shape (n, p)
            derivatives (jacobian) of the least squares fit for each parameter

        Returns
        -------
        freep_unc : array of shape (p,)
            uncertainties for each free paramater, in the same order as self.parameter_names
        """

        nparameters = deriv.shape[1]
        freep_unc = np.zeros(nparameters)

        for i in range(nparameters):
            pder = deriv[:, i]
            idx = pder != 0
            idx &= np.isfinite(pder)

            if np.count_nonzero(idx) <= 5:
                logger.warning(
                    "Not enough data points with a suitable derivative "
                    "to determine the uncertainties"
                )
                continue
            # Sort pixels according to the change of the i
            # parameter needed to match the observations
            idx_sort = np.argsort(resid[idx] / pder[idx])
            ch_x = resid[idx][idx_sort] / pder[idx][idx_sort]
            # Weights of the individual pixels also sorted
            # uncertainties are already included in pder / unc[idx][idx_sort]
            ch_y = np.abs(pder[idx][idx_sort])
            # Cumulative weights
            ch_y = np.cumsum(ch_y)
            # Normalized cumulative weights
            ch_y /= ch_y[-1]

            # hmed = np.interp(0.5, ch_y, ch_x)
            interval = np.interp([0.16, 0.84], ch_y, ch_x)
            sigma_estimate = (interval[1] - interval[0]) / 2
            freep_unc[i] = sigma_estimate

        return freep_unc

    def update_fitresults(self, sme, result, segments):
        # Update SME structure
        sme.fitresults.clear()

        popt = result.x
        sme.fitresults.values = popt
        sme.fitresults.parameters = self.parameter_names

        # Determine the covariance
        # hessian == fisher information matrix
        fisher = result.jac.T.dot(result.jac)
        covar = np.linalg.pinv(fisher)
        sig = np.sqrt(covar.diagonal())

        # Update fitresults
        sme.fitresults.covariance = covar
        sme.fitresults.gradient = result.grad
        sme.fitresults.derivative = result.jac
        sme.fitresults.residuals = (
            (sme.spec[segments] - sme.synth[segments]) / sme.uncs[segments]
        )[sme.mask_line[segments]]
        sme.fitresults.chisq = np.sum(sme.fitresults.residuals ** 2) / (
            result.fun.size - len(self.parameter_names)
        )
        sme.fitresults.iterations = self.iteration

        sme.fitresults.fit_uncertainties = [np.nan for _ in self.parameter_names]
        for i in range(len(self.parameter_names)):
            # Errors based on covariance matrix
            sme.fitresults.fit_uncertainties[i] = sig[i] * np.sqrt(sme.fitresults.chisq)

        try:
            sme.fitresults.uncertainties = self.estimate_uncertainties(
                sme.fitresults.residuals,
                sme.fitresults.derivative,
            )
        except:
            logger.warning(
                "Could not determine the uncertainties from the probability "
                "distribution, using fit uncertainties instead"
            )
            sme.fitresults.uncertainties = sme.fitresults.fit_uncertainties
        # sme.fitresults.uncertainties = sme.fitresults.fit_uncertainties

        return sme

    def sanitize_parameter_names(self, sme, param_names):
        # Sanitize parameter names
        param_names = [p.casefold() for p in param_names]
        param_names = [p.capitalize() if p[:5] == "abund" else p for p in param_names]

        param_names = [p if p != "grav" else "logg" for p in param_names]
        param_names = [p if p != "feh" else "monh" for p in param_names]

        # Parameters are unique
        # But keep the order the same
        param_names, index = np.unique(param_names, return_index=True)
        param_names = param_names[np.argsort(index)]
        param_names = list(param_names)

        if "vrad" in param_names:
            param_names.remove("vrad")
            if sme.vrad_flag in ["fix", "none"]:
                sme.vrad_flag = "whole"
                logger.info(
                    "Removed fit parameter 'vrad', instead set radial velocity flag to %s",
                    sme.vrad_flag,
                )

        if "cont" in param_names:
            param_names.remove("cont")
            if sme.cscale_flag in ["fix", "none"]:
                sme.cscale_flag = "linear"
                logger.info(
                    "Removed fit parameter 'cont', instead set continuum flag to %s",
                    sme.cscale_flag,
                )
        return param_names

    def solve(self, sme, param_names=None, segments="all", bounds=None):
        """
        Find the least squares fit parameters to an observed spectrum

        NOTE: intermediary results will be saved in filename ("sme.npy")

        Parameters
        ----------
        sme : SME_Struct
            sme struct containing all input (and output) parameters
        param_names : list, optional
            the names of the parameters to fit (default: ["teff", "logg", "monh"])
        filename : str, optional
            the sme structure will be saved to this file, use None to suppress this behaviour (default: "sme.npy")

        Returns
        -------
        sme : SME_Struct
            same sme structure with fit results in sme.fitresults, and best fit spectrum in sme.smod
        """

        assert "wave" in sme, "SME Structure has no wavelength"
        assert "spec" in sme, "SME Structure has no observation"

        if self.restore and self.filename is not None:
            fname = self.filename.rsplit(".", 1)[0]
            fname = f"{fname}_iter.json"
            try:
                with open(fname) as f:
                    data = json.load(f)
                for fp in param_names:
                    sme[fp] = data[fp]
                logger.warning(f"Restoring existing backup data from {fname}")
            except:
                pass

        if "uncs" not in sme:
            sme.uncs = np.ones(sme.spec.size)
            logger.warning("SME Structure has no uncertainties, using all ones instead")
        if "mask" not in sme:
            sme.mask = np.full(sme.wave.size, MASK_VALUES.LINE)

        segments = Synthesizer.check_segments(sme, segments)

        # Clean parameter values
        if param_names is None:
            param_names = sme.fitparameters
        if param_names is None or len(param_names) == 0:
            logger.warning(
                "No Fit Parameters have been set. Using ('teff', 'logg', 'monh') instead."
            )
            param_names = ("teff", "logg", "monh")
        self.parameter_names = self.sanitize_parameter_names(sme, param_names)

        self.update_linelist = False
        for name in self.parameter_names:
            if name[:8] == "linelist":
                if self.update_linelist is False:
                    self.update_linelist = []
                try:
                    idx = int(name.split()[1])
                except IndexError:
                    raise ValueError(
                        f"Could not parse fit parameter {name}, expected a "
                        "linelist parameter like 'linelist n gflog'"
                    )
                self.update_linelist += [idx]
        if self.update_linelist:
            self.update_linelist = np.unique(self.update_linelist)

        # Create appropiate bounds
        if bounds is None:
            bounds = self.get_bounds(sme)
        # scales = self.get_scale()
        step_sizes = self.get_step_sizes(self.parameter_names)
        # Starting values
        p0 = self.get_default_values(sme)
        if np.any((p0 < bounds[0]) | (p0 > bounds[1])):
            logger.warning(
                "Initial values are incompatible with the bounds, clipping initial values"
            )
            p0 = np.clip(p0, bounds[0], bounds[1])
        # Restore backup
        if self.restore:
            sme = self.restore_func(sme)

        # Get constant data from sme structure
        for seg in segments:
            sme.mask[seg, sme.uncs[seg] == 0] = MASK_VALUES.BAD
        mask = sme.mask_line[segments]
        spec = sme.spec[segments][mask]
        uncs = sme.uncs[segments][mask]

        # Divide the uncertainties by the spectrum, to improve the fit in the continuum
        # Just as in IDL SME, this increases the relative error for points inside lines
        # uncs /= np.abs(spec)

        # This is the expected range of the uncertainty
        # if the residuals are larger, they are dampened by log(1 + z)
        self.f_scale = 0.2 * np.nanmean(spec.ravel()) / np.nanmean(uncs.ravel())

        logger.info("Fitting Spectrum with Parameters: %s", ",".join(param_names))
        logger.debug("Initial values: %s", p0)
        logger.debug("Bounds: %s", bounds)

        if (
            sme.wran.min() * (1 - 100 / clight) > sme.linelist.wlcent.min()
            or sme.wran.max() * (1 + 100 / clight) < sme.linelist.wlcent.max()
        ):
            logger.warning(
                "The linelist extends far beyond the requested wavelength range."
                " This will slow down the calculation, consider using only relevant lines\n"
                f"Wavelength range: {sme.wran.min()} - {sme.wran.max()} Å"
                f" ; Linelist range: {sme.linelist.wlcent.min()} - {sme.linelist.wlcent.max()} Å"
            )

        # Setup LineList only once
        dll = self.synthesizer.get_dll()
        dll.SetLibraryPath()
        dll.InputLineList(sme.linelist)

        # Do the heavy lifting
        if self.nparam > 0:
            self.progressbar = tqdm(
                desc="Iteration", total=0, disable=~show_progress_bars
            )
            self.progressbar_jacobian = tqdm(
                desc="Jacobian", total=len(p0), disable=~show_progress_bars
            )
            with print_to_log():
                res = least_squares(
                    self._residuals,
                    jac=self._jacobian,
                    x0=p0,
                    bounds=bounds,
                    loss=sme.leastsquares_loss,
                    f_scale=self.f_scale,
                    method=sme.leastsquares_method,
                    x_scale=sme.leastsquares_xscale,
                    # These control the tolerance, for early termination
                    # since each iteration is quite expensive
                    xtol=sme.accxt,
                    ftol=sme.accft,
                    gtol=sme.accgt,
                    verbose=2,
                    args=(sme, spec, uncs, mask),
                    kwargs={
                        "bounds": bounds,
                        "segments": segments,
                        "step_sizes": step_sizes,
                        "method": sme.leastsquares_jac,
                    },
                )
                # The jacobian is altered by the loss function
                # This lets us keep the original for our uncertainty estimate
                res.jac = self._latest_jacobian

            self.progressbar.close()
            self.progressbar_jacobian.close()
            for i, name in enumerate(self.parameter_names):
                sme[name] = res.x[i]
            sme = self.update_fitresults(sme, res, segments)
            logger.debug("Reduced chi square: %.3f", sme.fitresults.chisq)
            try:
                for name, value, unc in zip(
                    self.parameter_names, res.x, sme.fitresults.fit_uncertainties
                ):
                    logger.info("%s\t%.5f +- %.5g", name.ljust(10), value, unc)
                logger.info("%s\t%s +- %s", "v_rad".ljust(10), sme.vrad, sme.vrad_unc)
            except:
                pass
        elif len(param_names) > 0:
            # This happens when vrad and/or cscale are given as parameters but nothing else
            # We could try to reuse the already calculated synthetic spectrum (if it already exists)
            # However it is much lower resolution then the newly synthethized one (usually)
            # Therefore the radial velocity wont be as good as when redoing the whole thing
            sme = self.synthesizer.synthesize_spectrum(sme, segments)
        else:
            raise ValueError("No fit parameters given")

        if self.filename is not None:
            sme.save(self.filename)

        return sme


def solve(
    sme, param_names=None, segments="all", filename=None, restore=False, **kwargs
):
    solver = SME_Solver(filename=filename, restore=restore)
    return solver.solve(sme, param_names, segments, **kwargs)



# -*- coding: utf-8 -*-
"""
Module for handling linelist data from the VALD3 database (http://vald.astro.uu.se/).


"""
import logging
import sys

import numpy as np
import pandas as pd
from astropy.io import fits

from .linelist import LineList, LineListError

logger = logging.getLogger(__name__)


class GesError(LineListError):
    """Vald Data File Error"""


class GesFile(LineList):
    """Atomic data for a list of spectral lines."""

    def __init__(self, filename, medium=None):
        self.filename = filename
        linelist = self.loads(filename)

        super().__init__(linelist, lineformat=self.lineformat, medium=self.medium)
        # Convert to desired medium
        if medium is not None:
            self.medium = medium

    @staticmethod
    def load(filename):
        """
        Read line data file from the VALD extract stellar service

        Parameters
        ----------
        filename : str
            Name of the VALD linelist file to read

        Returns
        -------
        vald : ValdFile
            Parsed vald file
        """
        return GesFile(filename)

    def loads(self, filename):
        logger.info("Loading GES file %s", filename)

        hdu = fits.open(filename)
        data = hdu[1].data

        linedata = {
            "species": data["name"][:, 0] + " " + data["ion"].astype(str),
            "atom_number": np.zeros(len(data)),
            "ionization": data["ion"],
            "wlcent": data["lambda"],
            "excit": data["e_low"],
            "gflog": data["log_gf"],
            "gflog_err": data["log_gf_err"],
            "gamrad": data["rad_damp"],
            "gamqst": data["stark_damp"],
            "gamvw": data["vdw_damp"],
            "lande": data["lande_mean"],
            "depth": data["depth"],
            "reference": data["lambda_ref"],
            "lande_lower": data["lande_low"],
            "lande_upper": data["lande_up"],
            "j_lo": data["j_low"],
            "e_upp": data["e_up"],
            "j_up": data["j_up"],
            "term_lower": data["label_low"],
            "term_upper": data["label_up"],
        }

        sort = np.argsort(linedata["wlcent"])

        # We need to make sure all data is in the system defined byteorder
        # Otherwise pandas will not work properly
        for key, value in linedata.items():
            if (value.dtype.byteorder == ">" and sys.byteorder == "little") or (
                value.dtype.byteorder == "<" and sys.byteorder == "big"
            ):
                value = value.byteswap().newbyteorder()
            linedata[key] = value[sort]

        linelist = pd.DataFrame.from_dict(linedata)
        self.lineformat = "long"
        self.unit = "Angstrom"
        self._medium = "air"

        # self.citation_info += self.parse_references(refdata, fmt)

        return linelist



# -*- coding: utf-8 -*-
"""
Provide Plotting utility for Jupyter Notebook using Plot.ly
Can also be used just for Plot.ly, which will then generated html files
"""
import json
import logging
from base64 import b64encode

import numpy as np
import plotly.graph_objs as go
import plotly.offline as py
from plotly.io import write_image
from scipy.constants import speed_of_light

from .plot_colors import PlotColors

try:
    import ipywidgets as widgets
    from IPython import get_ipython
    from IPython.display import display

    cfg = get_ipython()
    in_notebook = cfg is not None
except (AttributeError, ImportError, ModuleNotFoundError):
    in_notebook = False

logger = logging.getLogger(__name__)
clight = speed_of_light * 1e-3
fmt = PlotColors()

try:
    import htmlmin
except ImportError:
    logger.info("Install htmlmin for minified html files")
    htmlmin = None

if in_notebook:
    py.init_notebook_mode()


class FitPlot:
    """Plot the sme solve fit, as iterations pass along"""

    def __init__(self, wave, spec):
        self.fig = go.FigureWidget()
        self.fig.layout["xaxis"]["title"] = "Wavelength [Å]"
        self.fig.layout["yaxis"]["title"] = "Intensity"

        self.fig.add_scatter(x=wave, y=spec, name="Observation")

    def add_synth(self, wave, synth, iteration=0):
        """add a scatter plot to the plot"""
        self.fig.add_scatter(x=wave, y=synth, name=f"Iteration {iteration}")


class FinalPlot:
    """Big plot that covers everything"""

    def __init__(
        self,
        sme,
        segment=0,
        orig=None,
        labels=None,
    ):
        self.sme = sme
        self.wave = sme.wave
        self.spec = sme.spec
        self.mask = sme.mask
        self.smod = sme.synth
        self.orig = orig
        self.labels = labels if labels is not None else {}
        if sme.telluric is not None and self.smod is not None:
            for i in range(sme.nseg):
                if len(self.smod[i]) != 0:
                    self.smod[i] = self.smod[i] * sme.telluric[i]
        self.nsegments = len(self.wave)
        self.segment = segment
        self.wran = sme.wran
        self.lines = sme.linelist
        self.vrad = sme.vrad
        self.vrad = [v if v is not None else 0 for v in self.vrad]

        self.mask_type = "good"

        data, annotations = self.create_plot(self.segment)
        self.annotations = annotations

        # Add segment slider
        steps = []
        for i in range(self.nsegments):
            step = {
                "label": f"Segment {i}",
                "method": "update",
                "args": [
                    {"visible": [v == i for v in self.visible]},
                    {
                        "title": f"Segment {i}",
                        "annotations": annotations[i],
                        "xaxis": {
                            "range": list(self.wran[i]),
                            "linecolor": "black",
                            "linewidth": 2.4,
                            "mirror": True,
                            "ticks": "outside",
                            "tickcolor": "black",
                        },
                        "yaxis": {
                            "autorange": True,
                            "linecolor": "black",
                            "linewidth": 2.4,
                            "mirror": True,
                            "ticks": "outside",
                            "tickcolor": "black",
                        },
                    },
                ],
            }
            steps += [step]

        layout = {
            "dragmode": "select",
            "plot_bgcolor": "#fff",
            "selectdirection": "h",
            "title": f"Segment {self.segment}",
            "xaxis": {
                "title": "Wavelength [Å]",
                "linecolor": "black",
                "linewidth": 2.4,
                "mirror": True,
                "ticks": "outside",
                "tickcolor": "black",
            },
            "yaxis": {
                "title": "Intensity",
                "linecolor": "black",
                "linewidth": 2.4,
                "mirror": True,
                "ticks": "outside",
                "tickcolor": "black",
            },
            "annotations": annotations[self.segment],
            "sliders": [{"active": self.segment, "steps": steps}],
            "legend": {"traceorder": "reversed"},
        }
        if in_notebook:
            self.fig = go.FigureWidget(data=data, layout=layout)

            # add selection callback
            self.fig.data[0].on_selection(self.selection_fn)
        else:
            data = [go.Scattergl(d) for d in data]
            self.fig = go.FigureWidget(data=data, layout=layout)

        # Add button to save figure
        if in_notebook:
            self.button_save = widgets.Button(description="Save")
            self.button_save.on_click(self.save)

            # Add buttons for Mask selection
            self.button_mask = widgets.ToggleButtons(
                options=["Good", "Bad", "Continuum", "Line"],
                description="Mask",
            )
            self.button_mask.observe(self.on_toggle_click, "value")

            self.widget = widgets.VBox([self.button_mask, self.button_save, self.fig])
            display(self.widget)

    def save(self, _=None, filename="SME.html", **kwargs):
        """save plot to html file"""
        # Here we "hack" plotly to use base64 encoded data
        # since we have a lot of data to look at
        # this reduces the file size by about a factor 3
        # zipping it would provide another factor 2
        if filename.endswith(".html"):
            self.fig.layout.dragmode = "zoom"
            fig = self.fig.to_plotly_json()
            # Convert the data to base64
            for i in range(len(fig["data"])):
                for ax in ["x", "y"]:
                    data = np.asarray(fig["data"][i][ax]).astype("float32")
                    b64data = b64encode(data).decode("utf8")
                    fig["data"][i][ax] = {"data": b64data, "dtype": "float32"}

            data = json.dumps(fig["data"])
            layout = json.dumps(fig["layout"])
            # This is the entire html as generated by self.fig.to_html
            # except we replace the data and layout manually
            # and transform it from base64 to TypedArrays
            html = (
                r"""<html><head><meta charset="utf-8" /></head><body><div>"""
                r"""<script type="text/javascript">window.PlotlyConfig = {MathJaxConfig: 'local'};</script>"""
                r"""<script src="https://cdn.plot.ly/plotly-latest.min.js"></script>"""
                r"""<div id="56dedbe2-a786-4b0b-8232-b68dfd7b9eb5" class="plotly-graph-div" style="height:100%; width:100%;"></div>"""
                r"""<script type="text/javascript">"""
                r"""window.PLOTLYENV=window.PLOTLYENV || {};"""
                r"""if (document.getElementById("56dedbe2-a786-4b0b-8232-b68dfd7b9eb5")) {"""
                f"data = {data};"
                f"layout = {layout};"
                r"""const FromBase64 = function (str) {return new Uint8Array(atob(str).split('').map(function (c) { return c.charCodeAt(0); }));};"""
                r"""for (let i = 0; i < data.length; i++){"""
                r"""let data_x = data[i].x.data;"""
                r"""let data_y = data[i].y.data;"""
                r"""data[i].x = new Float32Array(FromBase64(data_x).buffer);"""
                r"""data[i].y = new Float32Array(FromBase64(data_y).buffer);"""
                r"""}"""
                r"""Plotly.newPlot("56dedbe2-a786-4b0b-8232-b68dfd7b9eb5", data, layout, {"responsive": true})};"""
                r"""</script></div></body></html>"""
            )
            # Finally try to minimize if possible
            if htmlmin is not None:
                html = htmlmin.minify(html, remove_empty_space=True)
            with open(filename, "w") as f:
                f.write(html)

        else:
            write_image(self.fig, filename)

    def show(self):
        self.fig.show()

    def shift_mask(self, x, mask):
        """shift the edges of the mask to the bottom of the plot,
        so that the mask creates a shape with straight edges"""
        for i in np.where(mask)[0]:
            try:
                if mask[i] == mask[i + 1]:
                    x[i] = x[i - 1]
                else:
                    x[i] = x[i + 1]
            except IndexError:
                pass

        return x

    def create_mask_points(self, x, y, mask, value):
        """Creates the points that define the outer edge of the mask region"""
        mask = mask != value
        x = np.copy(x)
        y = np.copy(y)
        y[mask] = 0
        x = self.shift_mask(x, mask)
        return x, y

    def create_plot(self, current_segment):
        """Generate the plot componentes (lines and masks) and line labels"""

        annotations = {}
        visible = []
        data = []
        line_mask_idx = {}
        cont_mask_idx = {}

        for seg in range(self.nsegments):

            k = len(visible)
            line_mask_idx[seg] = k
            cont_mask_idx[seg] = k + 1

            # The order of the plots is chosen by the z order, from low to high
            # Masks should be below the spectra (so they don't hide half of the line)
            # Synthetic on top of observation, because synthetic varies less than observation
            # Annoying I know, but plotly doesn't seem to have good controls for the z order
            # Or Legend order for that matter

            if (
                self.mask is not None
                and self.spec is not None
                and self.wave is not None
            ):
                # Line mask
                x, y = self.create_mask_points(
                    self.wave[seg], self.spec[seg], self.mask[seg], 1
                )

                data += [
                    dict(
                        x=x,
                        y=y,
                        fillcolor=fmt["LineMask"]["facecolor"],
                        fill="tozeroy",
                        mode="none",
                        name=self.labels.get("linemask", "Line Mask"),
                        hoverinfo="none",
                        legendgroup=2,
                        visible=current_segment == seg,
                    )
                ]
                visible += [seg]

                # Cont mask
                x, y = self.create_mask_points(
                    self.wave[seg], self.spec[seg], self.mask[seg], 2
                )

                data += [
                    dict(
                        x=x,
                        y=y,
                        fillcolor=fmt["ContMask"]["facecolor"],
                        fill="tozeroy",
                        mode="none",
                        name=self.labels.get("contmask", "Continuum Mask"),
                        hoverinfo="none",
                        legendgroup=2,
                        visible=current_segment == seg,
                    )
                ]
                visible += [seg]

            if self.spec is not None:
                # Observation
                data += [
                    dict(
                        x=self.wave[seg],
                        y=self.spec[seg],
                        line={"color": fmt["Obs"]["color"]},
                        name=self.labels.get("obs", "Observation"),
                        legendgroup=0,
                        visible=current_segment == seg,
                    )
                ]
                visible += [seg]

            # Synthetic, if available
            if self.smod is not None:
                data += [
                    dict(
                        x=self.wave[seg],
                        y=self.smod[seg],
                        name=self.labels.get("synth", "Synthetic"),
                        line={"color": fmt["Syn"]["color"]},
                        legendgroup=1,
                        visible=current_segment == seg,
                    )
                ]
                visible += [seg]
            if self.orig is not None:
                data += [
                    dict(
                        x=self.wave[seg],
                        y=self.orig[seg],
                        name=self.labels.get("orig", "Original"),
                        line={"color": fmt["Orig"]["color"], "dash": "dash"},
                        legendgroup=1,
                        visible=current_segment == seg,
                    )
                ]
                visible += [seg]

            # mark important lines
            if self.lines is not None and len(self.lines) != 0:
                seg_annotations = []
                xlimits = self.wave[seg][[0, -1]]
                xlimits *= 1 - self.vrad[seg] / clight
                lines = (self.lines.wlcent > xlimits[0]) & (
                    self.lines.wlcent < xlimits[1]
                )
                lines = self.lines[lines]

                # Filter out closely packaged lines of the same species
                # Threshold for the distance between lines
                wlcent, labels = [], []
                threshold = np.diff(xlimits)[0] / 100
                species = np.unique(lines["species"])
                for sp in species:
                    sp_lines = lines["wlcent"][lines["species"] == sp]
                    diff = np.diff(sp_lines)
                    sp_mask = diff < threshold
                    if np.any(sp_mask):
                        idx = np.where(np.diff(sp_mask))[0]
                        idx = [0, *idx, len(sp_mask) + 1]
                        idx = np.unique(idx)
                        for i, j in zip(idx[:-1], idx[1:]):
                            sp_wmid = np.mean(sp_lines[i:j])
                            sp_label = f"{sp} +{j-i-1}"
                            wlcent += [sp_wmid]
                            labels += [sp_label]
                    else:
                        for sp_wmid in sp_lines:
                            wlcent += [sp_wmid]
                            labels += [sp]

                wlcent = np.array(wlcent)
                labels = np.array(labels)

                # Keep only the 100 stongest lines for performance
                # if "depth" in lines.columns:
                #     lines.sort("depth", ascending=False)
                #     lines = lines[:20]
                # else:
                #     idx = np.random.choice(len(lines), 20, replace=False)
                #     lines = lines[idx]

                x = wlcent * (1 + self.vrad[seg] / clight)
                if self.smod is not None and len(self.smod[seg]) != 0:
                    y = np.interp(x, self.wave[seg], self.smod[seg])
                else:
                    y = np.interp(x, self.wave[seg], self.spec[seg])

                if self.smod is not None and len(self.smod[seg]) > 0:
                    ytop = np.max(self.smod[seg])
                elif self.spec is not None and len(self.spec[seg]) > 0:
                    ytop = np.max(self.spec[seg])
                else:
                    ytop = 1

                for i, line in enumerate(labels):
                    seg_annotations += [
                        {
                            "x": x[i],
                            "y": y[i],
                            "xref": "x",
                            "yref": "y",
                            "text": f"{labels[i]}",
                            "hovertext": f"{wlcent[i]}",
                            "textangle": 90,
                            "opacity": 1,
                            "ax": 0,
                            "ay": 1.2 * ytop,
                            "ayref": "y",
                            "showarrow": True,
                            "arrowhead": 7,
                            "xanchor": "left",
                        }
                    ]
                annotations[seg] = seg_annotations

        self.visible = visible
        self.line_mask_idx = line_mask_idx
        self.cont_mask_idx = cont_mask_idx

        return data, annotations

    def selection_fn(self, trace, points, selector):
        """Callback for area selection, changes the mask depending on selected mode"""
        self.segment = self.fig.layout["sliders"][0].active
        seg = self.segment

        xrange = selector.xrange
        wave = self.wave[seg]
        mask = self.mask[seg]

        # Choose pixels and value depending on selected type
        if self.mask_type == "good":
            value = 1
            idx = (wave > xrange[0]) & (wave < xrange[1]) & (mask == 0)
        elif self.mask_type == "bad":
            value = 0
            idx = (wave > xrange[0]) & (wave < xrange[1])
        elif self.mask_type == "line":
            value = 1
            idx = (wave > xrange[0]) & (wave < xrange[1]) & (mask != 0)
            print(np.count_nonzero(idx))
        elif self.mask_type == "cont":
            value = 2
            idx = (wave > xrange[0]) & (wave < xrange[1]) & (mask == 1)
        else:
            return

        # Apply changes if any
        if np.count_nonzero(idx) != 0:
            self.mask[seg][idx] = value

            with self.fig.batch_update():
                # Update Line Mask
                m = self.line_mask_idx[seg]
                x, y = self.create_mask_points(
                    self.wave[seg], self.spec[seg], self.mask[seg], 1
                )
                self.fig.data[m].x = x
                self.fig.data[m].y = y

                # Update Cont Mask
                m = self.cont_mask_idx[seg]
                x, y = self.create_mask_points(
                    self.wave[seg], self.spec[seg], self.mask[seg], 2
                )
                self.fig.data[m].x = x
                self.fig.data[m].y = y

    def on_toggle_click(self, change):
        """Callback for mask mode selector buttons"""
        change = change["new"]
        if change == "Good":
            self.set_mask_good()
        elif change == "Bad":
            self.set_mask_bad()
        elif change == "Continuum":
            self.set_mask_continuum()
        elif change == "Line":
            self.set_mask_line()

    def set_mask_good(self, _=None):
        """Called by clicking the 'good' mask button"""
        self.set_mask_type("good")

    def set_mask_bad(self, _=None):
        """Called by clicking the 'bad' mask button"""
        self.set_mask_type("bad")

    def set_mask_line(self, _=None):
        """Called by clicking the 'line' mask button"""
        self.set_mask_type("line")

    def set_mask_continuum(self, _=None):
        """Called by clicking the 'continuum' mask button"""
        self.set_mask_type("cont")

    def set_mask_type(self, type):
        """Changes the mask change mode and chooses the current interactive tool"""
        self.mask_type = type
        self.fig.layout["dragmode"] = "select"

    def add(self, x, y, label=""):
        """adds a scatter plot to the image, and makes the necessary changes in the slider"""
        self.fig.add_scatter(x=x, y=y, name=label, legendgroup=10)
        self.visible += [-1]

        # Update Sliders
        steps = []
        for i in range(self.nsegments):
            step_visible = [(v == i) or (v == -1) for v in self.visible]
            step = {
                "label": f"Segment {i}",
                "method": "update",
                "args": [
                    {"visible": step_visible},
                    {
                        "title": f"Segment {i}",
                        "annotations": self.annotations[i],
                        "xaxis": {"range": list(self.wran[i])},
                        "yaxis": {"autorange": True},
                    },
                ],
            }
            steps += [step]

        self.fig.layout["sliders"][0]["steps"] = steps



# -*- coding: utf-8 -*-
__file_ending__ = ".sme"

# Load correct version string
from ._version import get_versions

__version__ = get_versions()["version"]
del get_versions

# Add output to the console
import logging

import colorlog
import tqdm


class TqdmLoggingHandler(logging.Handler):
    def __init__(self, level=logging.NOTSET):
        super().__init__(level)

    def emit(self, record):
        try:
            msg = self.format(record)
            tqdm.tqdm.write(msg)
            self.flush()
        except (KeyboardInterrupt, SystemExit):
            raise
        except:
            self.handleError(record)


logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

console = TqdmLoggingHandler()
console.setLevel(logging.INFO)
console.setFormatter(
    colorlog.ColoredFormatter("%(log_color)s%(levelname)s - %(message)s")
)
logger.addHandler(console)

# Download library if it does not exist
import os.path
from ctypes import cdll

from .smelib import libtools

libfile = libtools.get_full_libfile()
if not os.path.exists(libfile):
    libtools.download_libsme()

try:
    cdll.LoadLibrary(libfile)
    from .smelib import _smelib
except:
    libtools.compile_interface()

# Provide submodules to the outside
__all__ = [
    "util",
    "abund",
    "atmosphere",
    "broadening",
    "continuum_and_radial_velocity",
    "cwrapper",
    "echelle",
    "iliffe_vector",
    "linelist",
    "nlte",
    "sme_synth",
    "sme",
    "solve",
    "uncertainties",
    "smelib",
]



# -*- coding: utf-8 -*-
"""
Handle the Json configuration file
At the moment it is only used for the LargeFileStorage
"""
import json
import logging
from os.path import dirname, exists, expanduser, join

logger = logging.getLogger(__name__)


def _requires_load(func):
    def func_new(self, *args, **kwargs):
        if self._cfg is None:
            self.load()
        return func(self, *args, **kwargs)

    return func_new


class Config:
    def __init__(self, fname="~/.sme/config.json"):
        self.filename = fname
        self._cfg = None

        if not exists(self.filename):
            logger.info(
                f"No cconfiguration file found at {self.filename}, using default values instead"
            )
            self.filename = join(dirname(__file__), "config_default.json")

    @property
    def filename(self):
        return str(self._filename)

    @filename.setter
    def filename(self, value):
        self._filename = expanduser(value)

    @_requires_load
    def __getitem__(self, key):
        return self._cfg[key]

    @_requires_load
    def __setitem__(self, key, value):
        self._cfg[key] = value

    def load(self):
        with open(self._filename, "r") as f:
            self._cfg = json.load(f)
        return self._cfg

    @_requires_load
    def save(self):
        with open(self._filename, "w") as f:
            json.dump(self._cfg, f)



# -*- coding: utf-8 -*-
from os.path import dirname, join

import numpy as np
import pytest

from pysme.abund import Abund
from pysme.linelist.linelist import LineList
from pysme.linelist.vald import ValdFile

species = "Fe 1"
wlcent = 5502.9931
excit = 0.9582
gflog = -3.047
gamrad = 7.19
gamqst = -6.22
gamvw = 239.249
linedata = [species, wlcent, excit, gflog, gamrad, gamqst, gamvw]


def test_line_init():
    """Test that property values equal line data passed to __init__()."""
    line = LineList(linedata)
    assert isinstance(line, LineList)
    assert line.species[0] == species
    assert line.wlcent[0] == wlcent
    assert line.excit[0] == excit
    assert line.gflog[0] == gflog
    assert line.gamrad[0] == gamrad
    assert line.gamqst[0] == gamqst
    assert line.gamvw[0] == gamvw


def test_linelist_add_and_len():
    """Test that len() returns the number of lines (including 0) in list."""
    linelist = LineList()
    assert isinstance(linelist, LineList)
    assert len(linelist) == 0
    for iline in range(3):
        print(len(linelist._lines))
        print(linelist._lines)
        assert len(linelist) == iline
        linelist.add(*linedata)


def test_linelist_properties():
    """Test that properties are lists with one item per line.
    Test that property value equal line data passed to add().
    """
    linelist = LineList()
    linelist.add(*linedata)
    proplist = [
        linelist.species,
        linelist.wlcent,
        linelist.excit,
        linelist.gflog,
        linelist.gamrad,
        linelist.gamqst,
        linelist.gamvw,
    ]
    for iprop, prop in enumerate(proplist):
        print(linelist._lines)
        assert isinstance(prop, np.ndarray)
        assert len(prop) == 1
        assert prop[0] == linedata[iprop]


def test_valdfile():
    """Test class to read a VALD line data file."""
    testdir = dirname(__file__)
    linelist = ValdFile(join(testdir, "testcase1.lin"))

    assert len(linelist) == 44
    assert linelist.lineformat == "short"
    assert linelist.medium == "air"
    assert linelist[0].species == "V 1"

    with pytest.raises(IOError):
        ValdFile(join(testdir, "testcase1.npy"))


def test_medium():
    """Test class to read a VALD line data file."""
    testdir = dirname(__file__)
    vf = ValdFile(join(testdir, "testcase1.lin"), medium="vac")

    assert vf.medium == "vac"


def test_short_format():
    linelist = ValdFile(join(dirname(__file__), "testcase1.lin"))

    assert linelist.lineformat == "short"
    assert len(linelist) == 44
    assert linelist.atomic is not None
    assert linelist.species is not None

    with pytest.raises(AttributeError):
        _ = linelist.lulande

    with pytest.raises(AttributeError):
        _ = linelist.extra

    assert isinstance(linelist.abund, Abund)
    assert isinstance(linelist.atmo, str)


def test_long_format():
    linelist = ValdFile(join(dirname(__file__), "testcase3.lin"))

    assert linelist.lineformat == "long"
    assert len(linelist) == 67
    assert linelist.atomic is not None
    assert linelist.species is not None
    assert linelist.lulande is not None
    assert linelist.extra is not None

    assert linelist.reference is not None
    assert linelist.error is not None
    assert linelist.term_lower is not None
    assert linelist.term_upper is not None

    assert isinstance(linelist.abund, Abund)
    assert isinstance(linelist.atmo, str)



# -*- coding: utf-8 -*-
from os import remove
from os.path import dirname

import numpy as np
import pytest

from pysme.sme import SME_Structure as SME_Struct


@pytest.fixture
def cwd():
    return dirname(__file__)


@pytest.fixture
def filename(cwd):
    fname = "{}/__test.sme".format(cwd)
    yield fname
    try:
        remove(fname)
    except:
        pass


def test_empty_structure():
    """Test that all properties behave well when nothing is set"""
    empty = SME_Struct()

    assert isinstance(empty.version, str)
    assert empty.teff is not None
    assert empty.logg is not None
    assert empty.vmic == 0
    assert empty.vmac == 0
    assert empty.vsini == 0

    assert empty.nseg == 0
    assert empty.wave is None
    assert empty.spec is None
    assert empty.uncs is None
    assert empty.synth is None
    assert empty.cont is None
    assert empty.mask is None
    assert empty.mask_good is None
    assert empty.mask_bad is None
    # assert empty.mask_line is None
    # assert empty.mask_continuum is None

    assert empty.cscale.shape == (0, 1)
    assert empty.vrad.shape == (0,)
    assert empty.cscale_flag == "none"
    assert empty.vrad_flag == "none"
    assert empty.cscale_degree == 0

    assert empty.mu is not None
    assert empty.nmu == 7

    # assert empty.md5 is not None

    assert empty.linelist is not None
    assert empty.species is not None
    assert len(empty.species) == 0
    assert empty.atomic is not None

    assert empty.monh == 0
    assert not np.isnan(empty["abund Fe"])
    assert empty.abund["H"] == 12
    assert not np.isnan(empty.abund()["Mg"])

    assert empty.system_info is not None
    assert empty.system_info.arch == ""

    assert len(empty.fitparameters) == 0
    assert empty.fitresults is not None
    assert empty.fitresults.covariance is None

    assert empty.atmo is not None
    assert empty.atmo.depth is None

    assert empty.nlte is not None
    assert empty.nlte.elements == []


def test_save_and_load_structure(filename):
    sme = SME_Struct()
    assert sme.teff is not None

    sme.teff = 5000
    sme.save(filename)
    del sme
    sme = SME_Struct.load(filename)
    assert sme.teff == 5000

    remove(filename)

    data = np.linspace(1000, 2000, 100)
    sme.wave = data
    sme.spec = data
    sme.save(filename)
    sme = SME_Struct.load(filename)
    assert np.all(sme.wave[0] == data)
    assert np.all(sme.spec[0] == data)
    assert sme.nseg == 1


def test_load_idl_savefile(cwd):
    filename = "{}/testcase1.inp".format(cwd)
    sme = SME_Struct.load(filename)

    assert sme.teff == 5770
    assert sme.wave is not None

    assert sme.nseg == 1
    assert sme.cscale_flag == "linear"
    assert sme.vrad_flag == "each"


def test_cscale_degree():
    sme = SME_Struct()
    sme.cscale = 1

    flags = ["none", "fix", "constant", "linear", "quadratic"]
    degrees = [0, 0, 0, 1, 2]

    for f, d in zip(flags, degrees):
        sme.cscale_flag = f
        assert sme.cscale_degree == d
        assert sme.cscale.shape[0] == 0
        assert sme.cscale.shape[1] == d + 1


def test_idlver():
    sme = SME_Struct()
    sme.system_info.update()
    # assert sme.idlver.arch == "x86_64"


def test_fitresults():
    sme = SME_Struct()
    sme.fitresults.chisq = 100
    sme.fitresults.clear()
    assert sme.fitresults.chisq is None



# -*- coding: utf-8 -*-
"""
NLTE module of SME
reads and interpolates departure coefficients from library files
"""

import itertools
import logging
import warnings

import numpy as np
from scipy import interpolate
from tqdm import tqdm

from .abund import Abund
from .abund import elements as abund_elem
from .data_structure import Collection, CollectionFactory, array, astype, oneof, this
from .util import show_progress_bars

logger = logging.getLogger(__name__)


class DirectAccessFile:
    """
    This function reads a single record from binary file that has the following
    structure:
    Version string              - 64 byte long
    # of directory bocks        - short int
    directory block length      - short int
    # of used directory blocks  - short int
    1st directory block
    key      - string of up to 256 characters padded with ' '
    datatype - 32-bit int 23-element array returned by SIZE
    pointer  - 64-bit int pointer to the beginning of the record

    2nd directory block
    ...
    last directory block
    1st data record
    ...
    last data record
    """

    def __init__(self, filename):
        key, pointer, dtype, shape, version = DirectAccessFile.read_header(filename)
        self.file = filename
        self.version = version
        self.key = key
        self.shape = shape
        self.pointer = pointer
        self.dtype = dtype

    def __getitem__(self, key):
        # Access data via brackets
        value = self.get(key)
        if value is None:
            raise KeyError(f"Key {key} not found")
        return value

    def get(self, key: str, alt=None) -> np.memmap:
        """get field from file"""
        idx = np.where(self.key == key)[0]

        if idx.size == 0:
            return alt
        else:
            idx = idx[0]

        return np.memmap(
            self.file,
            mode="r",
            offset=self.pointer[idx],
            dtype=self.dtype[idx],
            shape=self.shape[idx][::-1],
        )

    @staticmethod
    def idl_typecode(i):
        """
        relevant IDL typecodes and corresponding Numpy Codes
        Most specify a byte size, but not all
        """
        typecode = {
            0: "V",
            1: "B",
            2: "i2",
            3: "i4",
            4: "f4",
            5: "f8",
            6: "c4",
            7: "S",
            8: "O",
            9: "c8",
            10: "i8",
            11: "O",
            12: "u2",
            13: "u4",
            14: "i8",
            15: "u8",
        }
        return typecode[i]

    @staticmethod
    def get_typecode(dt):
        """
        relevant IDL typecodes and corresponding Numpy Codes
        Most specify a byte size, but not all
        """
        typecode = {
            "V": 0,
            "B": 1,
            "U": 1,
            "i2": 2,
            "i4": 3,
            "f4": 4,
            "f8": 5,
            "c4": 6,
            "S": 7,
            "O": 8,
            "c8": 9,
            "i8": 10,
            "O": 11,
            "u2": 12,
            "u4": 13,
            "i8": 14,
            "u8": 15,
        }
        if isinstance(dt, np.dtype):
            dt = dt.str
        if len(dt) > 2:
            dt = dt[1:]

        return typecode[dt]

    @staticmethod
    def get_dtypes(version):
        major, minor = version
        if major == 1 and minor == 00:
            header_dtype = np.dtype(
                [("nblocks", "<u2"), ("dir_length", "<u2"), ("ndir", "<u2")]
            )
            dir_dtype = np.dtype(
                [("key", "S256"), ("size", "<i4", 23), ("pointer", "<i8")]
            )
        elif major == 1 and minor >= 10:
            header_dtype = np.dtype(
                [("nblocks", "<u8"), ("dir_length", "<i2"), ("ndir", "<u8")]
            )
            dir_dtype = np.dtype(
                [("key", "S256"), ("size", "<i4", 23), ("pointer", "<i8")]
            )
        else:
            raise ValueError("DirectAccess File Version '{version}' not understood.")
        return header_dtype, dir_dtype

    @staticmethod
    def read_version(version_string: str):
        if isinstance(version_string, bytes):
            version_string = version_string.decode()

        major, minor = int(version_string[26]), int(version_string[28:30])
        version = (major, minor)
        return version

    @classmethod
    def read_header(cls, fname: str):
        """parse Header data"""
        with open(fname, "rb") as file:
            version_dtype = "S64"
            version = np.fromfile(file, version_dtype, count=1)
            version = cls.read_version(version[0])
            header_dtype, dir_dtype = cls.get_dtypes(version)
            header = np.fromfile(file, header_dtype, count=1)
            ndir = int(header["ndir"][0])
            directory = np.fromfile(file, dir_dtype, count=ndir)

        # Decode bytes to strings
        key = directory["key"]
        key = np.char.strip(np.char.decode(key))
        # Get relevant info from size parameter
        # ndim, n1, n2, ..., typecode, size
        dtype = np.array(
            [cls.idl_typecode(d[1 + d[0]]) for d in directory["size"]],
            dtype="U5",
        )
        shape = np.empty(ndir, dtype=object)
        shape[:] = [tuple(d[1 : d[0] + 1]) for d in directory["size"]]
        # Pointer to data arrays
        pointer = directory["pointer"]

        # Bytes (which represent strings) get special treatment to get the dimensions right
        # And to properly convert them to strings
        # Also Null terminator is important to remember
        idx = dtype == "B"
        dtype[idx] = [f"S{s[0]}" for s in shape[idx]]
        shape2 = np.empty(np.count_nonzero(idx), dtype=object)
        shape2[:] = [tuple(s[1:]) for s in shape[idx]]
        shape[idx] = shape2

        return key, pointer, dtype, shape, version

    @classmethod
    def write(cls, fname, **kwargs):
        major, minor = 1, 10
        ndir = len(kwargs)
        version = f"DirectAccess file Version {major}.{minor} 2011-03-24"
        version = np.asarray([version], dtype="S64")
        version_length = version.itemsize

        header_dtype, dir_dtype = cls.get_dtypes((major, minor))
        header_length = header_dtype.itemsize
        dir_length = dir_dtype.itemsize

        header = np.zeros(1, header_dtype)
        header[0]["nblocks"] = len(kwargs)
        header[0]["dir_length"] = dir_length
        header[0]["ndir"] = ndir

        directory = np.zeros(ndir, dir_dtype)
        pointer = version_length + header_length + ndir * dir_length

        for i, (key, value) in enumerate(kwargs.items()):
            value = np.asarray(value)

            directory[i]["key"] = key
            shape = directory[i]["size"]
            if np.issubdtype(value.dtype, np.dtype("U")) or np.issubdtype(
                value.dtype, np.dtype("S")
            ):
                value = value.astype("S")
                shape[0] = value.ndim + 1
                shape[1 : 2 + value.ndim] = value.itemsize, *value.shape
                shape[2 + value.ndim] = 1
                shape[3 + value.ndim] = value.size * value.itemsize
            else:
                shape[0] = value.ndim
                shape[1 : 1 + value.ndim] = value.shape
                shape[1 + value.ndim] = cls.get_typecode(value.dtype)
                shape[2 + value.ndim] = value.size

            directory[i]["pointer"] = pointer
            pointer += value.nbytes
            kwargs[key] = value

        with open(fname, "wb") as file:
            version.tofile(file)
            header.tofile(file)
            directory.tofile(file)

            for i, value in enumerate(kwargs.values()):
                value.tofile(file)


class Grid:
    """NLTE Grid class that handles all NLTE data reading and interpolation"""

    def __init__(
        self,
        sme,
        elem,
        lfs_nlte,
        selection="energy",
        solar=None,
        abund_format="Fe=12",
        min_energy_diff=None,
    ):
        #:str: Element of the NLTE grid
        self.elem = elem
        #:LineList: Whole LineList that was passed to the C library
        self.linelist = sme.linelist
        #:array(str): Elemental Species Names for the linelist
        self.species = sme.linelist.species
        #:str: Name of the grid
        self.grid_name = sme.nlte.grids[elem]
        #:str: complete filename of the NLTE grid data file
        self.fname = lfs_nlte.get(self.grid_name)
        #:{"levels", "energy"}: Selection algorithm to match lines in grid with linelist
        self.selection = selection
        #:float: Minimum difference between energy levels to match, only relevant for selection=='energy'
        self.min_energy_diff = min_energy_diff

        #:DirectAccessFile: The NLTE data file
        self.directory = DirectAccessFile(self.fname)
        self.version = self.directory.version

        # Check atmosphere compatibility
        if "atmosphere_grid" in self.directory.key:
            self.atmosphere_grid = self.directory["atmosphere_grid"][0]
            if self.atmosphere_grid != sme.atmo.source:
                logger.warning(
                    "The {elem} NLTE grid was created with {self.atmosphere_grid}, but we are using {sme.atmo.source} for the synthesis"
                )

        # The possible labels
        self._teff = self.directory["teff"]
        self._grav = self.directory["grav"]
        self._feh = self.directory["feh"]
        self._xfe = self.directory["abund"]
        # The position of the models in the datafile
        self._keys = self.directory["models"].astype("U")

        try:
            depth_name = str.lower(sme.atmo.interp)
        except TypeError:
            logger.warning(
                "No interpolation axis specified in the atmosphere, using 'rhox'."
            )
            depth_name = "rhox"

        try:
            depth = self.directory[depth_name]
        except KeyError:
            other_depth_name = "tau" if depth_name == "rhox" else "rhox"
            depth = self.directory[other_depth_name]
            logger.warning(
                f"No data for {depth_name} in NLTE grid for {self.elem} found, "
                f"using {other_depth_name} instead."
            )
            depth_name = other_depth_name

        #:str: Which parameter is used as the depth axis
        self.depth_name = depth_name
        #:array(float): depth points of the atmosphere that is passed to the C library (in log10 scale)
        self._depth = depth

        self._grid = None
        self._points = None

        #:list(int): number of points in the grid to cache for each parameter, order; abund, teff, logg, monh
        self.subgrid_size = sme.nlte.subgrid_size
        #:float: Solar Abundance of the element
        self.abund_format = abund_format
        if solar is None:
            solar = Abund.solar()
        elif isinstance(solar, Abund):
            pass
        else:
            solar = Abund(0, solar)
        self.solar = solar

        #:dict: upper and lower parameters covered by the grid
        self.limits = {}
        #:array: NLTE data array
        self.bgrid = None
        #:array: Depth points of the NLTE data
        self.depth = None

        #:array: Indices of the lines in the NLTE data
        self.linerefs = None
        #:array: Indices of the lines in the LineList
        self.lineindices = None
        #:array: Indices of the lines in the bgrid
        self.iused = None

        #:str: citations in bibtex format, if known
        self.citation_info = ""

        self.first_warning = True

        conf = self.directory["conf"].astype("U")
        term = self.directory["term"].astype("U")
        try:
            species = self.directory["spec"].astype("U")
        except KeyError:
            logger.warning(
                "Could not find 'species' field in NLTE file %s. Assuming they are all '%s 1'.",
                self.grid_name,
                self.elem,
            )
            species = np.full(conf.shape, "%s 1" % self.elem)
            pass
        rotnum = self.directory["J"]  # rotational number of the atomic state
        if self.version[0] == 1 and self.version[1] >= 10:
            energies = self.directory["energy"]  # energy in eV
            self.citation_info = self.directory["citation"][()].decode()
        else:
            self.citation_info = None
            if self.selection != "levels":
                # logger.warning(
                #     "NLTE grid file version %s only supports level selection, not %s",
                #     self.version,
                #     self.selection,
                # )
                self.selection = "levels"

        if self.selection == "levels":
            self.lineindices, self.linerefs, self.iused = self.select_levels(
                conf, term, species, rotnum
            )
        elif self.selection == "energy":
            self.lineindices, self.linerefs, self.iused = self.select_energies(
                conf, term, species, rotnum, energies
            )

    def solar_rel_abund(self, abund, elem):
        """Get the abundance of elem relative to H, i.e. [X/H]"""
        if self.abund_format == "H=12":
            # Its a bit more efficient, to simply use the values as is, instead of converting everything
            # just to end up back here
            idx = abund.elem_dict[elem]
            a = abund._pattern[idx]
            s = self.solar._pattern[idx]
            h = sh = 12
        else:
            a = abund.get_element(elem, type=self.abund_format)
            h = abund.get_element("H", type=self.abund_format)
            s = self.solar.get_element(elem, type=self.abund_format)
            sh = self.solar.get_element("H", type=self.abund_format)

        rabund = (a - h) - (s - sh)
        return rabund

    def scaled_rel_abund(self, abund):
        """Get the abundance of self.elem relative to Fe, i.e. [X/Fe]"""
        sel = self.solar_rel_abund(abund, self.elem)
        sfe = self.solar_rel_abund(abund, "Fe")
        rabund = sel - sfe
        return rabund

    def get(self, abund, teff, logg, monh, atmo):
        rabund = self.scaled_rel_abund(abund)

        if len(self.limits) == 0 or not (
            (self.limits["xfe"][0] <= rabund <= self.limits["xfe"][-1])
            and (self.limits["teff"][0] <= teff <= self.limits["teff"][-1])
            and (self.limits["grav"][0] <= logg <= self.limits["grav"][-1])
            and (self.limits["feh"][0] <= monh <= self.limits["feh"][-1])
        ):
            _ = self.read_grid(rabund, teff, logg, monh)

        return self.interpolate(rabund, teff, logg, monh, atmo)

    def read_grid(self, rabund, teff, logg, monh):
        """Read the NLTE coefficients from the nlte_grid files for the given element
        The class will cache subgrid_size points around the target values as well

        Parameters
        ----------
        rabund : float
            relative (to solar) abundance of the element
        teff : float
            temperature in Kelvin
        logg : float
            surface gravity in log(cgs)
        monh : float
            Metallicity in H=12

        Returns
        -------
        nlte_grid : dict
            collection of nlte coefficient data (memmapped)
        linerefs : array (nlines,)
            linelevel descriptions (Energy level terms)
        lineindices: array (nlines,)
            indices of the used lines in the linelist
        """
        # find the n nearest parameter values in the grid (n == subgrid_size)
        x = np.argsort(np.abs(rabund - self._xfe))[: self.subgrid_size[0]]
        t = np.argsort(np.abs(teff - self._teff))[: self.subgrid_size[1]]
        g = np.argsort(np.abs(logg - self._grav))[: self.subgrid_size[2]]
        f = np.argsort(np.abs(monh - self._feh))[: self.subgrid_size[3]]

        x = x[np.argsort(self._xfe[x])]
        t = t[np.argsort(self._teff[t])]
        g = g[np.argsort(self._grav[g])]
        f = f[np.argsort(self._feh[f])]

        # Read the models with those parameters, and store depth and level
        # Create storage array
        ndepths, nlevel = self.directory[self._keys[0, 0, 0, 0]].shape
        nabund = len(x)
        nteff = len(t)
        ngrav = len(g)
        nfeh = len(f)

        self.bgrid = np.zeros((ndepths, nlevel, nabund, nteff, ngrav, nfeh))

        for i, j, k, l in tqdm(
            np.ndindex(nabund, nteff, ngrav, nfeh),
            desc="Loading NLTE %s" % self.elem,
            total=nabund * nteff * ngrav * nfeh,
            disable=~show_progress_bars,
        ):
            model = self._keys[f[l], g[k], t[j], x[i]]
            try:
                self.bgrid[:, :, i, j, k, l] = self.directory[model]
            except KeyError:
                warnings.warn(
                    f"Missing Model for element {self.elem}: T={self._teff[t[j]]}, logg={self._grav[g[k]]}, feh={self._feh[f[l]]}, abund={self._xfe[x[i]]:.2f}"
                )
        mask = np.zeros(self._depth.shape[:-1], bool)
        for i, j, k in itertools.product(f, g, t):
            mask[i, j, k] = True
        self.depth = self._depth[mask, :]
        self.depth.shape = nfeh, ngrav, nteff, -1

        # Reduce the stored data to only relevant energy levels
        # Remap the previous indices into a collapsed sequence
        # level_labels = level_labels[iused]
        if self.iused is not None:
            self.bgrid = self.bgrid[self.iused, ...]
        else:
            self.bgrid = self.bgrid[False]

        self._points = (
            self._xfe[x],
            self._teff[t],
            self._grav[g],
            self._feh[f],
        )
        self.limits = {
            "teff": self._points[1][[0, -1]],
            "grav": self._points[2][[0, -1]],
            "feh": self._points[3][[0, -1]],
            "xfe": self._points[0][[0, -1]],
        }

        return self.bgrid

    def select_levels(self, conf, term, species, rotnum):
        """
        Match our NLTE terms to transitions in the vald3-format linelist.

        Level descriptions in the vald3 long format look like this:
        'LS                 2p6.3s               2S'
        'LS                 2p6.3p               2P*'
        These are stored in line_term_low and line_term_upp.
        The array line_extra has dimensions [3 x nlines]. It stores J_low, E_up, J_up
        The sme.atomic array stores:
        0) atomic number, 1) ionization state, 2) wavelength (in A),
        3) excitation energy of lower level (in eV), 4) log(gf), 5) radiative,
        6) Stark, 7) and van der Waals damping parameters

        Parameters
        ----------
        conf : array of shape (nl,)
            electronic configuration (for identification), e.g., 2p6.5s
        term : array of shape (nl,)
            term designations (for identification), e.g., a5F
        species : array of shape (nl,)
            Element and ion for each atomic level (for identification), e.g. Na 1.
        rotnum : array of shape (nl,)
            rotational number J of atomic state (for identification).

        Returns
        -------
        lineindices : array of shape (nl,)
            Indices of the used lines in the linelist
        linerefs : array of shape (2, nlines,)
            Cross references for the lower and upper level in each transition,
            to their indices in the list of atomic levels.
            Missing levels use indices of -1.
        iused : array of shape (nl,)
            Indices used in the subgrid
        """

        lineindices = np.asarray(self.species, "U")
        lineindices = np.char.startswith(lineindices, self.elem)
        if not np.any(lineindices):
            logger.warning(f"No NLTE transitions for {self.elem} found")
            return None, None, np.full(len(species), False)

        sme_species = self.species[lineindices]

        # Extract data from linelist
        low = self.linelist["term_lower"][lineindices]
        upp = self.linelist["term_upper"][lineindices]
        # Remove quotation marks (if any are there)
        low = low.astype(str)
        upp = upp.astype(str)
        low = np.char.replace(low, "'", "")
        upp = np.char.replace(upp, "'", "")
        # Get only the relevant part
        parts_low = np.char.partition(low, " ")[:, (0, 2)]
        parts_upp = np.char.partition(upp, " ")[:, (0, 2)]

        # Transform into term symbol J (2*S+1) ?
        extra = self.linelist.extra[lineindices]
        extra = extra[:, [0, 2]] * 2 + 1
        extra = np.rint(extra).astype("i8")

        # Transform into term symbol J (2*S+1) ?
        rotnum = np.rint(2 * rotnum + 1).astype(int)

        # Create record arrays for each set of labels
        dtype = [
            ("species", sme_species.dtype),
            ("configuration", parts_upp.dtype),
            ("term", parts_upp.dtype),
            ("J", extra.dtype),
        ]
        level_labels = np.rec.fromarrays((species, conf, term, rotnum), dtype=dtype)
        line_label_low = np.rec.fromarrays(
            (sme_species, parts_low[:, 0], parts_low[:, 1], extra[:, 0]),
            dtype=dtype,
        )
        line_label_upp = np.rec.fromarrays(
            (sme_species, parts_upp[:, 0], parts_upp[:, 1], extra[:, 1]),
            dtype=dtype,
        )

        # Prepare arrays
        nlines = parts_low.shape[0]
        linerefs = np.full((nlines, 2), -1)
        iused = np.zeros(len(species), dtype=bool)

        # Loop through the NLTE levels
        # and match line levels
        for i, level in enumerate(level_labels):
            idx_l = line_label_low == level
            linerefs[idx_l, 0] = i
            iused[i] |= np.any(idx_l)

            idx_u = line_label_upp == level
            linerefs[idx_u, 1] = i
            iused[i] |= np.any(idx_u)

        lineindices = np.where(lineindices)[0]

        # Remap the linelevel references
        for j, i in enumerate(np.where(iused)[0]):
            linerefs[linerefs == i] = j

        return lineindices, linerefs, iused

    def select_energies(self, conf, term, species, rotnum, energies):
        lineindices = np.asarray(self.species, "U")
        lineindices = np.char.startswith(lineindices, self.elem)
        if not np.any(lineindices):
            logger.warning("No NLTE transitions for %s found", self.elem)
            return None, None, np.full(len(species), False)
        sme_species = self.species[lineindices]

        # Extract data from linelist
        term_low = self.linelist["term_lower"][lineindices].astype(str)
        term_upp = self.linelist["term_upper"][lineindices].astype(str)
        # Remove quotation marks (if any are there)
        term_low = np.char.replace(term_low, "'", "")
        term_upp = np.char.replace(term_upp, "'", "")
        # Split the string into configuration and term
        term_low = np.char.partition(term_low, " ")
        term_upp = np.char.partition(term_upp, " ")
        # Remove whitespaces
        term_low = np.char.replace(term_low, " ", "")
        term_upp = np.char.replace(term_upp, " ", "")

        # Get only the relevant part
        conf_low, term_low = term_low[:, 0], term_low[:, 2]
        conf_upp, term_upp = term_upp[:, 0], term_upp[:, 2]

        # Energy levels in the linelist in eV
        energy_low = self.linelist["excit"][lineindices]
        energy_upp = self.linelist["e_upp"][lineindices]

        # Transform into term symbol J (2*S+1) ?
        extra = self.linelist.extra[lineindices]
        extra = extra[:, [0, 2]] * 2 + 1
        extra = np.rint(extra).astype(int)
        j_low, j_upp = extra[:, 0], extra[:, 1]

        # Transform into term symbol J (2*S+1) ?
        rotnum = np.rint(2 * rotnum + 1).astype(int)

        dtype = [
            ("species", sme_species.dtype),
            ("configuration", term_low.dtype),
            ("term", term_low.dtype),
            ("J", extra.dtype),
            ("energy", energy_low.dtype),
        ]

        level_labels = np.rec.fromarrays(
            (species, conf, term, rotnum, energies), dtype=dtype
        )
        line_label_low = np.rec.fromarrays(
            (sme_species, conf_low, term_low, j_low, energy_low), dtype=dtype
        )
        line_label_upp = np.rec.fromarrays(
            (sme_species, conf_upp, term_upp, j_upp, energy_upp), dtype=dtype
        )

        nlines = term_low.shape[0]  # np.count_nonzero(lineindices)
        linerefs = np.full((nlines, 2), -1)
        iused = np.zeros(len(species), dtype=bool)

        idx_map = np.arange(len(species))
        # Maximum energy in the grid
        max_energy = np.max(energies)
        # Maximum seperation between energy levels
        if self.min_energy_diff is None:
            diff = np.abs(np.diff(energies))
            energy_diff_limit = np.min(diff[diff != 0])
        else:
            energy_diff_limit = np.abs(self.min_energy_diff)

        def match(label, level):
            # Try to match the label and the level
            # Using various metrics in decreasing order of confidence
            idx_species = label.species == level.species
            idx_conf = label.configuration == level.configuration
            idx_term = label.term == level.term
            idx_j = label.J == level.J
            # 1. Try to match conf/term/spec/J as usual
            idx = idx_species & idx_conf & idx_term & idx_j
            if np.any(idx):
                return np.where(idx)[0][0]
            # 2. If it fails, try to match conf/term/spec, but ignore J
            # TODO: Unless there is another match!
            idx = idx_species & idx_conf & idx_term  # & ~idx_j
            if np.any(idx):
                return np.where(idx)[0][0]
            # 3. Try to match H
            # If spec is 'H 1', then try to match conf with conf,
            # *or* try to match term with term, *or* try to match conf with term,
            # *or* try to match term with conf
            if level.species == "H 1":
                # TODO
                idx_term_conf = label.term == level.configuration
                idx_conf_term = label.configuration == level.term
                idx = idx_conf | idx_term | idx_term_conf | idx_conf_term
                if np.any(idx):
                    return np.where(idx)[0][0]
            # 4. Try to match energies (including H, if step 3 failed)
            # Find the level in the nlte grid with the same spec,
            # and the closest energy; *provided* that the desired energy does
            # not exceed the smallest energy difference between levels in the grid
            idx = idx_species
            if np.any(idx):
                if level.energy > max_energy or level.energy <= 0:
                    # We exceed the maximum energy of the grid, ignore NLTE
                    return -1
                diff = np.abs(label[idx].energy - level.energy)
                idx2 = np.argmin(diff)
                mindiff = diff[idx2]
                # difference needs to be smaller than some limit?
                if mindiff < energy_diff_limit:
                    return np.where(idx_map[idx][idx2])[0][0]
                else:
                    return -1
            # 5. If everything fails return nothing
            return -1

        # Loop through the linelist line levels
        # and match to NLTE line levels
        # match() return -1 of no match is found, or the index otherwise
        for i, level in enumerate(line_label_low):
            idx_l = match(level_labels, level)
            linerefs[i, 0] = idx_l
            iused[idx_l] |= idx_l != -1

        for i, level in enumerate(line_label_upp):
            idx_u = match(level_labels, level)
            linerefs[i, 1] = idx_u
            iused[idx_u] |= idx_u != -1

        # Lineindices as integer pointers
        lineindices = np.where(lineindices)[0]
        # Remap the linelevel references
        for j, i in enumerate(np.where(iused)[0]):
            linerefs[linerefs == i] = j

        return lineindices, linerefs, iused

    def interpolate(self, rabund, teff, logg, monh, atmo):
        """
        interpolate nlte coefficients on the model grid

        Parameters
        ----------
        rabund : float
            relative (to solar) abundance of the element
        teff : float
            temperature in Kelvin
        logg : float
            surface gravity in log(cgs)
        monh : float
            Metallicity in H=12

        Returns
        -------
        subgrid : array (ndepth, nlines)
            interpolated grid values
        """

        assert self._points is not None
        if self.bgrid is None or self.bgrid.shape[0] == 0:
            return None

        # Interpolate the depth scale to the target depth, this is unstructured data
        # i.e. each combination of parameters has a different depth scale (given in depth)
        ndepths, _, *nparam = self.bgrid.shape
        target_depth = atmo[self.depth_name]
        target_depth = np.log10(target_depth)
        ntarget = len(target_depth)

        param = [rabund, teff, logg, monh]
        right = [
            (self._points[i][-1] == p) and len(self._points[i] > 1)
            for i, p in enumerate(param)
        ]
        iabund = np.digitize(rabund, self._points[0], right=right[0]) - 1
        iteff = np.digitize(teff, self._points[1], right=right[1]) - 1
        ilogg = np.digitize(logg, self._points[2], right=right[2]) - 1
        imonh = np.digitize(monh, self._points[3], right=right[3]) - 1

        # Interpolate on the grid
        # We only interpolate on the 2**4 points around the requested values
        # self._points and self._grid are interpolated when reading the data in read_grid
        target = (rabund, teff, logg, monh)
        points = (
            self._points[0][iabund : iabund + 2],
            self._points[1][iteff : iteff + 2],
            self._points[2][ilogg : ilogg + 2],
            self._points[3][imonh : imonh + 2],
        )
        npoints = (
            max(points[0].size, 1),
            max(points[1].size, 1),
            max(points[2].size, 1),
            max(points[3].size, 1),
        )

        grid = np.empty((*npoints, ntarget, ndepths), float)
        for l, x, t, g, f in np.ndindex(ndepths, *npoints):
            xp = self.depth[imonh + f, ilogg + g, iteff + t, :]
            yp = self.bgrid[l, :, iabund + x, iteff + t, ilogg + g, imonh + f]
            xp = np.log10(xp)
            grid[x, t, g, f, :, l] = interpolate.interp1d(
                xp,
                yp,
                bounds_error=False,
                fill_value="extrapolate",
                kind="cubic",
            )(target_depth)

        # Check if we need to extrapolate
        if any(
            [t < min(p) or t > max(p) for t, p in zip(target, points) if len(p) > 0]
        ):
            if self.first_warning:
                logger.warning(
                    f"Extrapolate on the {self.elem} NLTE grid. Requested values of {target} on grid {points}"
                )
                self.first_warning = False

        # Some grids have only one value in that direction
        # Usually in abundance. Then we need to remove that dimension
        # to avoid nan output
        mask = [len(p) > 1 for p in points]
        if not all(mask):
            points = [p for m, p in zip(mask, points) if m]
            target = [t for m, t in zip(mask, target) if m]
            idx = [slice(None, None) if m else 0 for m in mask]
            grid = grid[tuple(idx)]

        method = "order"
        if method == "grid":
            # TODO: Interpolate with splines
            # Possibly in order of importance, since scipy doesn't have spline interpolation on a grid
            subgrid = interpolate.interpn(
                points,
                grid,
                target,
                method="linear",
                bounds_error=False,
                fill_value=None,
            )
            subgrid = subgrid[0]
        elif method == "order":
            for p, t in zip(points, target):
                grid = interpolate.interp1d(
                    p,
                    grid,
                    axis=0,
                    bounds_error=False,
                    fill_value="extrapolate",
                )(t)
            subgrid = grid
        return subgrid


@CollectionFactory
class NLTE(Collection):
    # fmt: off
    _fields = Collection._fields + [
        ("elements", [], astype(list), this,
            "list: elements for which nlte calculations will be performed"),
        ("grids", {}, astype(dict), this,
            "dict: nlte grid datafiles for each element"),
        ("subgrid_size", [3, 3, 3, 3], array(4, int), this,
            "array of shape (4,): defines size of nlte grid cache."
            "Each entry is for one parameter abund, teff, logg, monh"),
        ("flags", None, array(None, np.bool_), this,
            "array: contains a flag for each line, whether it was calculated in NLTE (True) or not (False)"),
        ("solar", None, this, this, "str: defines which default to use as the solar metallicitiies"),
        ("abund_format", "H=12", astype(str), this, "str: which abundance format to use for comparison"),
        ("selection", "energy", oneof("energy", "levels"), this, "str: which selection algorithm to use to match linelist and departure coefficients"),
        ("min_energy_diff", None, this, this, "float: difference between energy levels that are still matched. If None will default to the smallest non zero difference between energy levels in the grid.")
    ]
    # fmt: on

    # For marcs2012 atmosphere
    _default_grids = {
        "Al": "nlte_Al_ama51_pysme.grd",
        "Fe": "marcs2012_Fe2016.grd",
        "Li": "nlte_Li_ama51_pysme.grd",
        "Mg": "nlte_Mg_ama51_pysme.grd",
        "Na": "nlte_Na_ama51_pysme.grd",
        "O": "nlte_O_ama51_pysme.grd",
        "Ba": "nlte_Ba_ama51_pysme.grd",
        "Ca": "nlte_Ca_ama51_pysme.grd",
        "Si": "nlte_Si_ama51_pysme.grd",
        "Ti": "marcs2012s_t2.0_Ti.grd",
        "C": "nlte_C_ama51_pysme.grd",
        "H": "nlte_H_ama51_pysme.grd",
        "K": "nlte_K_ama51_pysme.grd",
        "Mn": "nlte_Mn_ama51_pysme.grd",
        "N": "nlte_N_ama51_pysme.grd",
    }

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.first = False

        if "solar" in kwargs.keys():
            self.solar = kwargs["solar"]

        # Convert IDL keywords to Python
        if "nlte_elem_flags" in kwargs.keys():
            elements = kwargs["nlte_elem_flags"]
            self.elements = [abund_elem[i] for i, j in enumerate(elements) if j == 1]

        if "nlte_subgrid_size" in kwargs.keys():
            self.subgrid_size = kwargs["nlte_subgrid_size"]

        if "nlte_grids" in kwargs:
            grids = kwargs["nlte_grids"]
            if isinstance(grids, (list, np.ndarray)):
                grids = {
                    abund_elem[i]: name.decode()
                    for i, name in enumerate(grids)
                    if name != ""
                }
            self.grids = grids

        # TODO
        #:dict: the cached subgrid data for each element
        # This is NOT saved on sme.save
        # But maybe should be ?
        self.grid_data = {}

    def set_nlte(self, element, grid=None):
        """
        Add an element to the NLTE calculations

        Parameters
        ----------
        element : str
            The abbreviation of the element to add to the NLTE calculations
        grid : str, optional
            Filename of the NLTE data grid to use for this element
            the file must be in nlte_grids directory
            Defaults to a set of "known" files for some elements
        """
        if element in self.elements:
            # Element already in NLTE
            # Change grid if given
            if grid is not None:
                self.grids[element] = grid
            return

        if grid is None:
            # Use default grid
            if element not in NLTE._default_grids.keys():
                raise ValueError(f"No default grid known for element {element}")
            grid = NLTE._default_grids[element]
            logger.info("Using default grid %s for element %s", grid, element)

        if element not in self.elements:
            self.elements += [element]
        self.grids[element] = grid

    def remove_nlte(self, element):
        """
        Remove an element from the NLTE calculations

        Parameters
        ----------
        element : str
            Abbreviation of the element to remove from NLTE
        """
        if element not in self.elements:
            # Element not included in NLTE anyways
            return

        self.elements.remove(element)
        self.grids.pop(element)

    @property
    def _citation_info(self):
        citations = [
            grid.citation_info
            for el, grid in self.grid_data.items()
            if grid.citation_info is not None
        ]
        citations = "\n".join(citations)
        return citations

    @_citation_info.setter
    def _citation_info(self, value):
        pass

    def update_coefficients(self, sme, dll, lfs_nlte):
        """pass departure coefficients to C library"""

        # Only print "Running in NLTE" message on the first run each time
        if np.all(self.grids == "") or np.size(self.elements) == 0:
            # No NLTE to do
            if self.first:
                self.first = False
                logger.info("Running in LTE")
            return sme
        if sme.linelist.lineformat == "short":
            if self.first:
                self.first = False
                logger.warning(
                    "NLTE line formation was requested, but VALD3 long-format linedata\n"
                    "are required in order to relate line terms to NLTE level corrections!\n"
                    "Line formation will proceed under LTE."
                )
            return sme

        # Reset the departure coefficient every time, just to be sure
        # It would be more efficient to just Update the values, but this doesn't take long
        dll.ResetDepartureCoefficients()

        if self.first:
            self.first = False
            logger.info("Running in NLTE: %s", ", ".join(self.elements))

        # Call each element to update and return its set of departure coefficients
        marked_for_removal = []
        for elem in self.elements:
            # Call function to retrieve interpolated NLTE departure coefficients
            # the abundances for NLTE are handled in the H-12 format
            grid = self.get_grid(sme, elem, lfs_nlte)
            if not np.any(grid.iused):
                # No lines are found for this element
                # remove it from the elements after this loop
                logger.warning(
                    "No %s NLTE lines found, removing it from NLTE calculations",
                    elem,
                )
                marked_for_removal += [elem]
                continue
            bmat = grid.get(sme.abund, sme.teff, sme.logg, sme.monh, sme.atmo)
            if bmat is None or grid.linerefs.size == 0:
                logger.warning(
                    "No %s NLTE lines found, removing it from NLTE calculations",
                    elem,
                )
                marked_for_removal += [elem]
                continue
            else:
                # Put corrections into the nlte_b matrix, don't cache the data
                for lr, li in zip(grid.linerefs, grid.lineindices):
                    # loop through the list of relevant _lines_, substitute both their levels into the main b matrix
                    # Make sure both levels have corrections available
                    if lr[0] != -1 and lr[1] != -1:
                        dll.InputDepartureCoefficients(bmat[:, lr], li)

        for elem in marked_for_removal:
            self.remove_nlte(elem)

        # flags = sme_synth.GetNLTEflags(sme.linelist)

        return sme

    def get_grid(self, sme, elem, lfs_nlte):
        """Read and interpolate the NLTE grid for the current element and parameters"""
        if self.grids[elem] is None:
            raise ValueError(f"Element {elem} has not been prepared for NLTE")

        # The grids are cached in the NLTE object, but not saved
        if elem in self.grid_data.keys():
            grid = self.grid_data[elem]
        else:
            grid = Grid(
                sme,
                elem,
                lfs_nlte,
                solar=self.solar,
                abund_format=self.abund_format,
                selection=self.selection,
                min_energy_diff=self.min_energy_diff,
            )
            self.grid_data[elem] = grid

        return grid


# These are kept here for compatibility, but it is recommended to use the functions of sme.nlte
def nlte(sme, dll, elem, lfs_nlte):
    """Read and interpolate the NLTE grid for the current element and parameters"""
    grid = sme.nlte.get_grid(sme, elem, lfs_nlte)
    subgrid = grid.get(sme.abund, sme.teff, sme.logg, sme.monh, sme.atmo)
    return subgrid, grid.linerefs, grid.lineindices


def update_nlte_coefficients(sme, dll, lfs_nlte):
    """pass departure coefficients to C library"""
    return sme.nlte.update_coefficients(sme, dll, lfs_nlte)



# -*- coding: utf-8 -*-
""" Handles reading and interpolation of atmopshere (grid) data """
import logging

import numpy as np
from flex.extensions.bindata import MultipleDataExtension

from ..abund import Abund
from ..data_structure import (
    Collection,
    CollectionFactory,
    absolute,
    array,
    asfloat,
    asstr,
    lowercase,
    oneof,
    this,
    uppercase,
)

logger = logging.getLogger(__name__)


class AtmosphereError(RuntimeError):
    """Something went wrong with the atmosphere interpolation"""


@CollectionFactory
class Atmosphere(Collection):
    """
    Atmosphere structure
    contains all information to describe the solar atmosphere
    i.e. temperature etc in the different layers
    as well as stellar parameters and abundances
    """

    # fmt: off
    _fields = [
        ("teff", 5770, asfloat, this, "float: effective temperature in Kelvin"),
        ("logg", 4.0, asfloat, this, "float: surface gravity in log10(cgs)"),
        ("abund", Abund.solar(), this, this, "Abund: elemental abundances"),
        ("vturb", 0, absolute, this, "float: turbulence velocity in km/s"),
        ("lonh", 0, asfloat, this, "float: ?"),
        ("source", "marcs2012.sav", this, this, "str: datafile name of this data, or atmosphere grid/file"),
        ("method", "grid", lowercase(oneof("grid", "embedded")), this,
            "str: whether the data source is a grid or a fixed atmosphere"),
        ("geom", "PP", uppercase(oneof("PP", "SPH", None)), this,
            "str: the geometry of the atmopshere model"),
        ("radius", 0, asfloat, this, "float: radius of the spherical model"),
        ("height", None, array(None, "f8"), this, "array: height of the spherical model"),
        ("opflag", [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1], array(20, int), this,
            "array of size (20,): opacity flags"),
        ("wlstd", 5000, asfloat, this, "float: wavelength standard deviation"),
        ("depth", None, uppercase(oneof(None, "RHOX", "TAU")), this,
            "str: flag that determines whether to use RHOX or TAU for calculations"),
        ("interp", None, uppercase(oneof(None, "RHOX", "TAU")), this,
            "str: flag that determines whether to use RHOX or TAU for interpolation"),
        ("rhox", None, array(None, "f8"), this,
            "array: mass column density"),
        ("tau", None, array(None, "f8"), this,
            "array: continuum optical depth"),
        ("temp", None, array(None, "f8"), this,
            "array: temperature profile in Kelvin"),
        ("rho", None, array(None, "f8"), this,
            "array: density profile"),
        ("xna", None, array(None, "f8"), this,
            "array: number density of atoms in 1/cm**3"),
        ("xne", None, array(None, "f8"), this,
            "array: number density of electrons in 1/cm**3"),
        ("citation_info", "", asstr, this, "str: citation for this atmosphere"),
    ]
    # fmt: on

    # TODO: pick the right geometry for the grid, based on whether it has height/radius values or not
    def __init__(self, *args, **kwargs):
        monh = kwargs.pop("monh", kwargs.pop("feh", 0))
        abund = kwargs.pop("abund", "empty")
        abund_format = kwargs.pop("abund_format", "sme")
        if "geom" in kwargs.keys() and kwargs["geom"] == "":
            kwargs["geom"] = None
        super().__init__(**kwargs)
        self.abund = Abund(monh=monh, pattern=abund, type=abund_format)

    @property
    def monh(self):
        """float: metallicity"""
        return self.abund.monh

    @monh.setter
    def monh(self, value):
        self.abund.monh = value

    @property
    def names(self):
        return self._names + ["monh"]

    @property
    def dtype(self):
        obj = lambda: None
        obj.names = [n.lower() for n in self.names]
        return obj

    @property
    def ndep(self):
        scale = self.temp
        if scale is not None:
            return scale.shape[0]
        return None

    @ndep.setter
    def ndep(self, value):
        pass

    def _save(self):
        data = {}
        ext2 = self.abund._save()
        header = ext2.header
        data["abund"] = ext2.data
        header["abund_format"] = header["type"]
        del header["type"]

        for name in self._names:
            value = self[name]

            if isinstance(value, np.ndarray):
                data[name] = value
            elif value is None or isinstance(value, Abund):
                pass
            else:
                # if isinstance(value, (np.floating, np.integer, np.str, float, int, str)):
                header[name] = value

        ext = MultipleDataExtension(header, data)

        return ext

    @classmethod
    def _load(cls, ext):
        header = ext.header
        header.update(ext.data)
        obj = cls(**header)
        return obj


@CollectionFactory
class AtmosphereGrid(np.recarray):
    """
    A grid of atmospheres, used for the interpolation
    of model atmospheres. Each entry represents one
    atmosphere model.
    """

    # fmt: off
    _fields = [
        ("source", None, asstr, this, "str: datafile name of this data"),
        ("method", "grid", lowercase(oneof("grid", "embedded")), this,
            "str: whether the data source is a grid or a fixed atmosphere"),
        ("geom", None, uppercase(oneof(None, "PP", "SPH")), this,
            "str: the geometry of the atmopshere model"),
        ("depth", None, uppercase(oneof(None, "RHOX", "TAU")), this,
            "str: flag that determines whether to use RHOX or TAU for calculations"),
        ("interp", None, uppercase(oneof(None, "RHOX", "TAU")), this,
            "str: flag that determines whether to use RHOX or TAU for interpolation"),
        ("abund_format", "sme", oneof(*Abund._formats), this,
            "str: format of the Abundance field, as defined by the Abund class"),
        ("citation_info", "", asstr, this, "str: Citation text to cite in your papers"),
    ]
    # fmt: on

    def __new__(cls, natmo, npoints, **kwargs):
        dtype = [
            ("teff", "f4"),
            ("logg", "f4"),
            ("monh", "f4"),
            ("vturb", "f4"),
            ("lonh", "f4"),
            ("radius", "f4"),
            ("height", f"({npoints},)f4"),
            ("wlstd", "f4"),
            ("opflag", "(20,)i4"),
            ("abund", "(99,)f4"),
            ("temp", f"({npoints},)f4"),
            ("rhox", f"({npoints},)f4"),
            ("tau", f"({npoints},)f4"),
            ("rho", f"({npoints},)f4"),
            ("xna", f"({npoints},)f4"),
            ("xne", f"({npoints},)f4"),
        ]

        names = [s[0].lower() for s in dtype]
        titles = [s[0].upper() for s in dtype]

        data = super().__new__(cls, (natmo,), dtype=dtype, names=names, titles=titles)

        data.interp = "TAU"
        data.depth = "RHOX"
        data.method = "grid"
        data.geom = "PP"
        data.source = ""
        data.citation_info = ""
        data.abund_format = "sme"
        data.info = ""
        return data

    def __array_finalize__(self, obj):
        if obj is None:
            return
        self.interp = getattr(obj, "interp", "TAU")
        self.depth = getattr(obj, "depth", "RHOX")
        self.method = getattr(obj, "method", "grid")
        self.geom = getattr(obj, "geom", "PP")
        self.source = getattr(obj, "source", "")
        self.citation_info = getattr(obj, "citation_info", "")
        self.abund_format = getattr(obj, "abund_format", "sme")
        self.info = getattr(obj, "info", "")

    def __reduce__(self):
        # Get the parent's __reduce__ tuple
        pickled_state = super(AtmosphereGrid, self).__reduce__()
        # Create our own tuple to pass to __setstate__
        new_state = pickled_state[2] + (
            self.interp,
            self.depth,
            self.source,
            self.geom,
            self.citation_info,
            self.method,
            self.abund_format,
            self.info,
        )
        # Return a tuple that replaces the parent's __setstate__ tuple with our own
        return (pickled_state[0], pickled_state[1], new_state)

    def __setstate__(self, state):
        self.interp = state[-8]
        self.depth = state[-7]
        self.source = state[-6]
        self.geom = state[-5]
        self.citation_info = state[-4]
        self.method = state[-3]
        self.abund_format = state[-2]
        self.info = state[-1]

        # Call the parent's __setstate__ with the other tuple elements.
        super(AtmosphereGrid, self).__setstate__(state[0:-8])

    def __getitem__(self, key):
        """Overwrite the getitem routine, so we keep additional
        properties and/or return an atmosphere object, when only
        one record is returned"""
        cls = self.__class__
        value = super().__getitem__(key)
        if isinstance(value, cls) and value.size == 1:
            return value[0]

        if isinstance(value, np.record):
            kwargs = {s: value[s] for s in value.dtype.names}
            value = Atmosphere(**kwargs)
        if isinstance(value, (Atmosphere, cls)):
            for name in self._names:
                setattr(value, name, getattr(self, name))
        return value

    def __str__(self):
        return self.source

    def __repr__(self):
        return str(self)

    def get(self, teff, logg, monh):
        mask = self.teff == teff
        mask &= self.logg == logg
        mask &= self.monh == monh
        return self[mask]

    @property
    def ndep(self):
        return self.shape[1]

    def save(self, filename):
        """Save the Atmopshere grid to a file using a numpy save format"""
        header = {
            "interp": self.interp,
            "depth": self.depth,
            "source": self.source,
            "geom": self.geom,
            "citation_info": self.citation_info,
            "method": self.method,
            "abund_format": self.abund_format,
            "info": self.info,
        }
        names = list(header.keys())
        values = [header[n] for n in names]
        header = np.rec.array(values, names=names)
        np.savez(filename, data=self, header=header)
        return

    @classmethod
    def load(cls, filename):
        """Load the atmosphere grid data from disk"""
        data = np.load(filename)
        self = data["data"].view(cls)
        header = data["header"]
        for k in header.dtype.names:
            v = header[k][()]
            setattr(self, k, v)
        return self



# -*- coding: utf-8 -*-
import logging
import os.path
import platform
import sys
from datetime import datetime as dt
from enum import Enum, IntEnum, IntFlag

import numpy as np
from scipy.constants import speed_of_light
from scipy.io import readsav

from . import __file_ending__, __version__, echelle, persistence
from .abund import Abund
from .abund import elements as abund_elem
from .atmosphere.atmosphere import Atmosphere
from .data_structure import *
from .iliffe_vector import Iliffe_vector
from .linelist.linelist import LineList
from .nlte import NLTE

logger = logging.getLogger(__name__)


class MASK_VALUES(IntFlag):
    """
    Mask value specifier used in mob

    Values can be combined to mark a point as e.g. line and vrad
    If cont or vrad are not used it will fallback to the line mask
    """

    BAD = 0
    LINE = 1
    CONT = 2
    VRAD = 4


class CONT_SCALE(Enum):
    NONE = "none"
    FIX = "fix"
    LINEAR = "linear"
    QUADRATIC = "quadratic"
    CUBIC = "cubic"


class CONT_OPTIONS(Enum):
    MASK = "mask"
    MATCH = "match"
    MATCH_MASK = "match+mask"
    MATCHLINES = "matchlines"
    MATCHLINES_MASK = "matchlines+mask"
    SPLINE = "spline"
    SPLINE_MASK = "spline_mask"
    MCMC = "mcmc"


class VRAD_OPTIONS(Enum):
    NONE = "none"
    FIX = "fix"
    EACH = "each"
    WHOLE = "whole"


@CollectionFactory
class Parameters(Collection):
    # fmt: off
    _fields = Collection._fields + [
        ("teff", 5770, asfloat, this, "float: effective temperature in Kelvin"),
        ("logg", 4.0, asfloat, this, "float: surface gravity in log10(cgs)"),
        ("abund", Abund.solar(), this, this, "Abund: elemental abundances"),
        ("vmic", 0, absolute, this, "float: micro turbulence in km/s"),
        ("vmac", 0, absolute, this, "float: macro turbulence in km/s"),
        ("vsini", 0, absolute, this, "float: projected rotational velocity in km/s"),
    ]
    # fmt: on

    def __init__(self, **kwargs):
        monh = kwargs.pop("monh", kwargs.pop("feh", 0))
        abund = kwargs.pop("abund", "solar")
        if "grav" in kwargs.keys() and "logg" not in kwargs.keys():
            kwargs["logg"] = kwargs.pop("grav")
        super().__init__(**kwargs)
        self.abund = Abund(monh=monh, pattern=abund, type="sme")

    @property
    def _abund(self):
        return self.__abund

    @_abund.setter
    def _abund(self, value):
        if isinstance(value, Abund):
            self.__abund = value
        else:
            logger.warning(
                "Abundance set using just a pattern, assuming that "
                "it has format %s. "
                "If that is incorrect, try changing the format first.",
                self.__abund.type,
            )
            self.__abund = Abund(monh=self.monh, pattern=value, type=self.__abund.type)

    @property
    def monh(self):
        """float: metallicity in log scale relative to the base abundances"""
        return self.abund.monh

    @monh.setter
    def monh(self, value):
        self.abund.monh = value

    def citation(self, format="string"):
        return self.abund.citation()


@CollectionFactory
class Version(Collection):
    # fmt: off
    _fields = Collection._fields + [
        ("arch", "", asstr, this, "str: system architecture"),
        ("os", "", asstr, this, "str: operating system"),
        ("os_family", "", asstr, this, "str: operating system family"),
        ("os_name", "", asstr, this, "str: os name"),
        ("release", "", asstr, this, "str: python version"),
        ("build_date", "", asstr, this, "str: build date of the Python version used"),
        ("memory_bits", 64, astype(int), this, "int: Platform architecture bit size (usually 32 or 64)"),
        ("host", "", asstr, this, "str: name of the machine that created the SME Structure")
    ]
    # fmt: on

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def update(self):
        """Update version info with current machine data"""
        self.arch = platform.machine()
        self.os = sys.platform
        self.os_family = platform.system()
        self.os_name = platform.version()
        self.release = platform.python_version()
        self.build_date = platform.python_build()[1]
        self.memory_bits = int(platform.architecture()[0][:2])
        self.host = platform.node()


@CollectionFactory
class Fitresults(Collection):
    # fmt: off
    _fields = Collection._fields + [
        ("iterations", None, astype(int, allow_None=True), this, "int: maximum number of iterations in the solver"),
        ("chisq", None, this, this, "float: reduced chi-square of the solution"),
        ("parameters", None, this, this, "list: parameter names"),
        ("values", None, array(None, float), this, "array: best fit values for the fit parameters"),
        ("uncertainties", None, array(None, float), this, "array of size(nfree,): uncertainties of the free parameters bases on SME statistics"),
        ("covariance", None, array(None, float), this, "array of size (nfree, nfree): covariance matrix"),
        ("gradient", None, array(None, float), this, "array of size (nfree,): final gradients of the free parameters on the cost function"),
        ("derivative", None, array(None, float), this, "array of size (npoints, nfree): final Jacobian of each point and each parameter"),
        ("residuals", None, array(None, float), this, "array of size (npoints,): final residuals of the fit"),
        ("fit_uncertainties", None, this, this, "array: uncertainties based solely on the least_squares fit")
    ]
    # fmt: on

    def clear(self):
        """Reset all values to None"""
        for name in self._names:
            default = [f[1] for f in self._fields if f[0] == name][0]
            setattr(self, name, default)


@CollectionFactory
class SME_Structure(Parameters):
    # fmt: off
    _fields = Parameters._fields + [
        ("id", dt.now(), asstr, this, "str: DateTime when this structure was created"),
        ("meta", {}, this, this, "dict: Arbitrary extra information"),
        ("version", __version__, this, this, "str: PySME version used to create this structure"),
        ("vrad_flag", "none", lowercase(oneof(-2, -1, 0, "none", "each", "whole", "fix")), this,
            """str: flag that determines how the radial velocity is determined

            allowed values are:
               * "none": No radial velocity correction
               * "each": Determine radial velocity for each segment individually
               * "whole": Determine one radial velocity for the whole spectrum
            """),
        ("vrad", 0, array(None, float), this, "array of size (nseg,): radial velocity of each segment in km/s"),
        ("vrad_bounds", (-500, 500), array(2, float), this, "float: radial velocity limits in km/s"),
        ("vrad_loss", "soft_l1", asstr, this, "str: loss function for the radial velocity fit"),
        ("vrad_method", "dogbox", asstr, this, "str: least squares method used in the radial velocity fit"),
        ("vrad_jac", "3-point", asstr, this, "str: jacobian approximation used in the radial velocity fit"),
        ("vrad_xscale", "jac", this, this, "array or 'jac': scale of the vrad parameter"),
        ("vrad_ftol", 1e-8, asfloat, this, "float: tolerance for the radial velocity least squares fit"),
        ("vrad_xtol", 1e-8, asfloat, this, "float: tolerance for the radial velocity least squares fit"),
        ("vrad_gtol", 1e-8, asfloat, this, "float: tolerance for the radial velocity least squares fit"),
        ("cscale_flag", "none", lowercase(oneof("none", "fix", "constant", "linear", "quadratic", "cubic", "quintic", "quantic", astype=int)), this,
            """str: Flag that describes how to correct for the continuum

            allowed values are:
                * "none": No continuum correction
                * "fix": Use whatever continuum scale has been set, but don't change it
                * "constant": Zeroth order polynomial, i.e. scale everything by a factor
                * "linear": First order polynomial, i.e. approximate continuum by a straight line
                * "quadratic": Second order polynomial, i.e. approximate continuum by a quadratic polynomial
            """),
        ("cscale_type", "match+mask", lowercase(oneof("mcmc", "mask", "match", "match+mask", "matchlines", "matchlines+mask", "spline", "spline+mask")), this,
            """str: Flag that determines the algorithm to determine the continuum

            This is used in combination with cscale_flag, which determines the degree of the fit, if any.

            allowed values are:
              * "whole": Fit the whole synthetic spectrum to the observation to determine the best fit
              * "mask": Fit a polynomial to the pixels marked as continuum in the mask
            """),
        ("cscale", None, this, this,
            """array of size (nseg, ndegree): Continumm polynomial coefficients for each wavelength segment
            The x coordinates of each polynomial are chosen so that x = 0, at the first wavelength point,
            i.e. x is shifted by wave[segment][0]
            """),
        ("cscale_bounds", (-np.inf, np.inf), this, this, "array(2, cscale_degree): bounds for the continuum parameters"),
        ("cscale_loss", "soft_l1", asstr, this, "str: loss function for the continuum fit"),
        ("cscale_method", "dogbox", asstr, this, "str: least squares method used in the continuum fit"),
        ("cscale_jac", "3-point", asstr, this, "str: jacobian approximation used in the continuum fit"),
        ("cscale_xscale", "jac", this, this, "array of 'jac', Scale of each continuum parameter"),
        ("cscale_ftol", 1e-8, asfloat, this, "float: tolerance for the continuum least squares fit"),
        ("cscale_xtol", 1e-8, asfloat, this, "float: tolerance for the continuum least squares fit"),
        ("cscale_gtol", 1e-8, asfloat, this, "float: tolerance for the continuum least squares fit"),
        ("normalize_by_continuum", True, asbool, this,
            "bool: Whether to normalize the synthetic spectrum by the synthetic continuum spectrum or not"),
        ("specific_intensities_only", False, asbool, this,
            "bool: Whether to keep the specific intensities or integrate them together"),
        ("gam6", 1, asfloat, this, "float: van der Waals scaling factor"),
        ("h2broad", True, asbool, this, "bool: Whether to use H2 broadening or not"),
        ("accwi", 3e-3, asfloat, this,
            "float: minimum accuracy for linear spectrum interpolation vs. wavelength."),
        ("accrt", 1e-4, asfloat, this,
            "float: minimum accuracy for synthethized spectrum at wavelength grid points in sme.wint."),
        ("leastsquares_method", "dogbox", asstr, this, "str: leastsquares method to use, see scipy least_squares for details, default: 'dogbox'."),
        ("leastsquares_loss", "linear", asstr, this, "str: leastsquares loss to use, see scipy least_squares for details, default: 'linear'"),
        ("leastsquares_xscale", 1.0, this, this, "str, arraylike: leastsquare x-scale to use, see scipy least_squares for details, default: 1"),
        ("leastsquares_jac", "2-point", asstr, this, "str: leastsquares jacobian calculation, see scipy least_squares for details, default: '2-point'"),
        ("leastsquares_ftol", 1e-3, asfloat, this, "float: minimum accuracy of the best fit cost"),
        ("leastsquares_xtol", 1e-6, asfloat, this, "float: minimum accuracy of the parameters in the fitting procedure"),
        ("leastsquares_gtol", 1e-4, asfloat, this, "float: minimum accuracy of the gradient of the least squares fit"),
        ("iptype", None, lowercase(oneof(None, "gauss", "sinc", "table")), this, "str: instrumental broadening type"),
        ("ipres", 0, array(None, float), this, "float, array: Instrumental resolution for instrumental broadening"),
        ("ip_x", None, this, this, "array: Instrumental broadening table in x direction"),
        ("ip_y", None, this, this, "array: Instrumental broadening table in y direction"),
        ("mu", np.sqrt(0.5 * (2 * np.arange(7) + 1) / 7), array(None, float), this,
            """array of size (nmu,): Mu values to calculate radiative transfer at
            mu values describe the distance from the center of the stellar disk to the edge
            with mu = cos(theta), where theta is the angle of the observation,
            i.e. mu = 1 at the center of the disk and 0 at the edge
            """),
        ("wran", None, this, this,
            "array of size (nseg, 2): beginning and end wavelength points of each segment"),
        ("wave", None, vector, this,
            "Iliffe_vector of shape (nseg, ...): wavelength"),
        ("spec", None, vector, this,
            "Iliffe_vector of shape (nseg, ...): observed spectrum"),
        ("uncs", None, vector, this,
            "Iliffe_vector of shape (nseg, ...): uncertainties of the observed spectrum"),
        ("telluric", None, vector, this,
            "Illife_vector of shape (nseg, ...): telluric spectrum that is multiplied with synth during the fit"),
        ("mask", None, vector, this,
            "Iliffe_vector of shape (nseg, ...): mask defining good and bad points for the fit"),
        ("synth", None, vector, this,
            "Iliffe_vector of shape (nseg, ...): synthetic spectrum"),
        ("cont", None, vector, this,
            "Iliffe_vector of shape (nseg, ...): continuum intensities"),
        ("linelist", LineList(), astype(LineList), this, "LineList: spectral line information"),
        ("fitparameters", [], astype(list, allow_None=True), this, "list: parameters to fit"),
        ("fitresults", Fitresults(), astype(Fitresults), this, "Fitresults: fit results data"),
        ("atmo", Atmosphere(), astype(Atmosphere), this, "Atmosphere: model atmosphere data"),
        ("nlte", NLTE(), astype(NLTE), this, "NLTE: nlte calculation data"),
        ("system_info", Version(), astype(Version), this,
            "Version: information about the host system running the calculation for debugging")
    ]
    # fmt: on

    def __init__(self, **kwargs):
        wind = kwargs.get("wind", None)

        atmo = kwargs.pop("atmo", {})
        nlte = kwargs.pop("nlte", {})
        idlver = kwargs.pop("idlver", {})
        self.wave = None
        self.wran = None
        super().__init__(**kwargs)

        if wind is not None and self.wave is not None:
            wind = wind + 1
            self.wave = Iliffe_vector(self.wave.ravel(), offsets=wind)

        self.spec = kwargs.get("sob", None)
        self.uncs = kwargs.get("uob", None)
        self.mask = kwargs.get("mob", None)
        self.synth = kwargs.get("smod", None)

        self.meta["object"] = kwargs.get("obs_name", "")
        try:
            self.linelist = LineList(**kwargs)
        except (KeyError, AttributeError):
            # TODO ignore the warning during loading of data
            logger.warning("No or incomplete linelist data present")

        # Parse free parameters into one list
        pname = kwargs.get("pname", [])
        glob_free = kwargs.get("glob_free", [])
        if isinstance(glob_free, str):
            glob_free = [glob_free]
        ab_free = kwargs.get("ab_free", [])
        if len(ab_free) != 0:
            ab_free = [f"abund {el}" for i, el in zip(ab_free, abund_elem) if i == 1]
        fitparameters = np.concatenate((pname, glob_free, ab_free)).astype("U")
        #:array of size (nfree): Names of the free parameters
        self.fitparameters = np.unique(fitparameters)

        self.fitresults = Fitresults(
            iterations=kwargs.get("maxiter", None),
            chisq=kwargs.get("chisq", 0),
            uncertainties=kwargs.get("punc", None),
            covariance=kwargs.get("covar", None),
        )

        self.normalize_by_continuum = kwargs.get("cscale_flag", "") != "fix"

        self.system_info = Version(**idlver)
        atmo_abund = atmo.pop("abund", kwargs.get("abund", "empty"))
        atmo_monh = atmo.pop(
            "monh", atmo.pop("feh", kwargs.get("monh", kwargs.get("feh", 0)))
        )
        self.atmo = Atmosphere(**atmo, abund=atmo_abund, monh=atmo_monh)
        self.nlte = NLTE(**nlte)

        self.citation_info = r"""
            @ARTICLE{2017A&A...597A..16P,
                author = {{Piskunov}, Nikolai and {Valenti}, Jeff A.},
                title = "{Spectroscopy Made Easy: Evolution}",
                journal = {\aap},
                keywords = {stars: abundances, radiative transfer, stars: fundamental parameters, stars: atmospheres, techniques: spectroscopic, Astrophysics - Instrumentation and Methods for Astrophysics, Astrophysics - Solar and Stellar Astrophysics},
                year = "2017",
                month = "Jan",
                volume = {597},
                eid = {A16},
                pages = {A16},
                doi = {10.1051/0004-6361/201629124},
                archivePrefix = {arXiv},
                eprint = {1606.06073},
                primaryClass = {astro-ph.IM},
                adsurl = {https://ui.adsabs.harvard.edu/abs/2017A&A...597A..16P},
                adsnote = {Provided by the SAO/NASA Astrophysics Data System}
            }
            @ARTICLE{1996A&AS..118..595V,
                author = {{Valenti}, J.~A. and {Piskunov}, N.},
                title = "{Spectroscopy made easy: A new tool for fitting observations with synthetic spectra.}",
                journal = {\aaps},
                keywords = {RADIATIVE TRANSFER, METHODS: NUMERICAL, TECHNIQUES: SPECTROSCOPIC, STARS: FUNDAMENTAL PARAMETERS, SUN: FUNDAMENTAL PARAMETERS, ATOMIC DATA},
                year = "1996",
                month = "Sep",
                volume = {118},
                pages = {595-603},
                adsurl = {https://ui.adsabs.harvard.edu/abs/1996A&AS..118..595V},
                adsnote = {Provided by the SAO/NASA Astrophysics Data System}
            }
            """

        # Apply final conversions from IDL to Python version
        if "wave" in self:
            self.__convert_cscale__()

    def __getitem__(self, key):
        assert isinstance(key, str), "Key must be of type string"
        key = key.casefold()

        if key.startswith("abund "):
            element = key[5:].strip()
            element = element.capitalize()
            return self.abund[element]
        if key.startswith("linelist "):
            _, idx, field = key[8:].split(" ", 2)
            idx = int(idx)
            return self.linelist[field][idx]
        return super().__getitem__(key)

    def __setitem__(self, key, value):
        assert isinstance(key, str), "Key must be of type string"
        key = key.casefold()

        if key.startswith("abund "):
            element = key[5:].strip()
            element = element.capitalize()
            self.abund[element] = value
        elif key.startswith("linelist "):
            _, idx, field = key[8:].split(" ", 2)
            idx = int(idx)
            self.linelist[field][idx] = value
        else:
            super().__setitem__(key, value)

    # Additional constraints on fields
    @property
    def _wran(self):
        if self.wave is not None:
            nseg = self.wave.shape[0]
            values = np.zeros((nseg, 2))
            for i in range(nseg):
                if self.wave[i] is not None and len(self.wave[i]) >= 2:
                    values[i] = [self.wave[i][0], self.wave[i][-1]]
                else:
                    values[i] = self.__wran[i]
            self.__wran = values
            return self.__wran
        return self.__wran

    @_wran.setter
    def _wran(self, value):
        try:
            if self.wave is not None:
                logger.warning(
                    "The wavelength range is overriden by the existing wavelength grid"
                )
        except:
            pass
        self.__wran = np.atleast_2d(value) if value is not None else None

    @property
    def _vrad(self):
        """array of size (nseg,): Radial velocity in km/s for each wavelength region"""
        nseg = self.nseg if self.nseg is not None else 1

        if self.__vrad is None:
            self.__vrad = np.zeros(nseg)
            return self.__vrad

        if self.vrad_flag == "none":
            return np.zeros(nseg)
        else:
            nseg = self.__vrad.shape[0]
            if nseg == self.nseg:
                return self.__vrad

            rv = np.zeros(self.nseg)
            rv[:nseg] = self.__vrad[:nseg]
            rv[nseg:] = self.__vrad[-1]
            self.__vrad = rv
            return self.__vrad

        return self.__vrad

    @_vrad.setter
    def _vrad(self, value):
        self.__vrad = np.atleast_1d(value) if value is not None else None

    @property
    def _vrad_flag(self):
        try:
            _ = self.__vrad_flag
        except AttributeError:
            self.__vrad_flag = "none"
        return self.__vrad_flag

    @_vrad_flag.setter
    def _vrad_flag(self, value):
        if isinstance(value, (int, np.integer)):
            value = {-2: "none", -1: "whole", 0: "each"}[value]
        self.__vrad_flag = value

    @property
    def _cscale(self):
        """array of size (nseg, ndegree): Continumm polynomial coefficients for each wavelength segment
        The x coordinates of each polynomial are chosen so that x = 0, at the first wavelength point,
        i.e. x is shifted by wave[segment][0]
        """
        if self.cscale_type in ["spline", "spline+mask"]:
            return self.__cscale

        nseg = self.nseg if self.nseg is not None else 1

        if self.__cscale is None:
            cs = np.zeros((nseg, self.cscale_degree + 1))
            cs[:, -1] = 1
            self.__cscale = cs
            return cs

        if self.cscale_flag == "none":
            return np.ones((nseg, 1))

        ndeg = self.cscale_degree + 1
        ns, nd = self.__cscale.shape

        if nd == ndeg and ns == nseg:
            return self.__cscale

        cs = np.zeros((nseg, ndeg))
        cs[:, -1] = 1
        if nd > ndeg:
            cs[:ns, :] = self.__cscale[:ns, -ndeg:]
        elif nd < ndeg:
            cs[:ns, -nd:] = self.__cscale[:ns, :]
        else:
            cs[:ns, :] = self.__cscale[:ns, :]

        cs[ns:, -1] = 1

        # We need to update our internal representation as well
        # since we might operate on that array
        self.__cscale = cs

        return self.__cscale

    @_cscale.setter
    def _cscale(self, value):
        if self.cscale_type in ["spline", "spline+mask"]:
            if not isinstance(value, Iliffe_vector):
                self.__cscale = vector(self, value)
            else:
                self.__cscale = value
        else:
            self.__cscale = np.atleast_2d(value) if value is not None else None

    @property
    def _cscale_flag(self):
        try:
            _ = self.__cscale_flag
        except AttributeError:
            self.__cscale_flag = "none"
        return self.__cscale_flag

    @_cscale_flag.setter
    def _cscale_flag(self, value):
        if isinstance(value, (int, np.integer)):
            try:
                value = {
                    -3: "none",
                    -2: "fix",
                    -1: "fix",
                    0: "constant",
                    1: "linear",
                    2: "quadratic",
                    3: "cubic",
                    4: "quintic",
                    5: "quantic",
                }[value]
            except KeyError:
                value = value

        self.__cscale_flag = value

    @property
    def _mu(self):
        return self.__mu

    @_mu.setter
    def _mu(self, value):
        if np.any(value < 0):
            raise ValueError("All values must be positive")
        if np.any(value > 1):
            raise ValueError("All values must be smaller or equal to 1")
        # The mu values are expected in decreasing order
        # For spherical models
        value = np.sort(value)[::-1]
        self.__mu = value

    @property
    def _ipres(self):
        return self.__ipres

    @_ipres.setter
    def _ipres(self, value):
        size = np.size(value)
        if self.nseg != 0 and size != 1 and size != self.nseg:
            raise ValueError(
                f"The instrument resolution must have 1 or {self.nseg} elements"
            )
        self.__ipres = value

    # Additional properties
    @property
    def nseg(self):
        """int: Number of wavelength segments"""
        if self.wran is None:
            return 0
        else:
            return len(self.wran)

    @property
    def nmu(self):
        return self.mu.size

    @nmu.setter
    def nmu(self, value):
        self.mu = np.sqrt(0.5 * (2 * np.arange(value) + 1) / value)

    @property
    def mask_good(self):
        if self.mask is None:
            return None
        return self.mask != MASK_VALUES.BAD

    @property
    def mask_bad(self):
        if self.mask is None:
            return None
        return self.mask == MASK_VALUES.BAD

    @property
    def mask_line(self):
        if self.mask is None:
            return None
        return (self.mask & MASK_VALUES.LINE) != 0

    @property
    def mask_cont(self):
        if self.mask is None:
            return None
        return (self.mask & MASK_VALUES.CONT) != 0

    @property
    def mask_vrad(self):
        if self.mask is None:
            return None
        return (self.mask & MASK_VALUES.VRAD) != 0

    @property
    def cscale_degree(self):
        """int: Polynomial degree of the continuum as determined by cscale_flag"""
        if self.cscale_type in ["spline", "spline+mask"]:
            return self.wave.shape[1]
        else:
            if self.cscale_flag == "constant":
                return 0
            if self.cscale_flag == "linear":
                return 1
            if self.cscale_flag == "quadratic":
                return 2
            if self.cscale_flag == "cubic":
                return 3
            if self.cscale_flag == "quintic":
                return 4
            if self.cscale_flag == "quantic":
                return 5
            if self.cscale_flag == "fix":
                # Use the underying element to avoid a loop
                if self.__cscale is not None:
                    return self.__cscale.shape[1] - 1
                else:
                    return 0
            if self.cscale_flag == "none":
                return 0
            return self.cscale_flag
            raise ValueError("This should never happen")

    @property
    def atomic(self):
        """array of size (nlines, 8): Atomic linelist data, usually passed to the C library
        Use sme.linelist instead for other purposes"""
        if self.linelist is None:
            return None
        return self.linelist.atomic

    @property
    def species(self):
        """array of size (nlines,): Names of the species of each spectral line"""
        if self.linelist is None:
            return None
        return self.linelist.species

    # Aliases for outdated names
    @property
    def accxt(self):
        return self.leastsquares_xtol

    @accxt.setter
    def accxt(self, value):
        self.leastsquares_xtol = value

    @property
    def accft(self):
        return self.leastsquares_ftol

    @accft.setter
    def accft(self, value):
        self.leastsquares_ftol = value

    @property
    def accgt(self):
        return self.leastsquares_gtol

    @accgt.setter
    def accgt(self, value):
        self.leastsquares_gtol = value

    @property
    def vrad_limit(self):
        return self.vrad_bounds[0]

    @vrad_limit.setter
    def vrad_limit(self, value):
        self.vrad_bounds = (-value, value)

    def __convert_cscale__(self):
        """
        Convert IDL SME continuum scale to regular polynomial coefficients
        Uses Taylor series approximation, as IDL version used the inverse of the continuum
        """
        wave = self.wave
        self.cscale = np.require(self.cscale, requirements="W")

        if self.cscale_flag == "linear":
            for i in range(len(self.cscale)):
                c, d = self.cscale[i]
                a, b = max(wave[i]), min(wave[i])
                c0 = (a - b) * (c - d) / (a * c - b * d) ** 2
                c1 = (a - b) / (a * c - b * d)

                # Shift zero point to first wavelength of the segment
                c1 += c0 * self.spec[i][0]

                self.cscale[i] = [c0, c1]
        elif self.cscale_flag == "fix":
            self.cscale = self.cscale / np.sqrt(2)
        elif self.cscale_flag == "constant":
            self.cscale = np.sqrt(1 / self.cscale)

    def import_mask(self, other, keep_bpm=False):
        """
        Import the mask of another sme structure and apply it to this one
        Conversion is based on the wavelength

        Parameters
        ----------
        other : SME_Structure
            the sme structure to import the mask from

        Returns
        -------
        self : SME_Structure
            this sme structure
        """
        if self.mask is None:
            self.mask = MASK_VALUES.LINE

        c_light = speed_of_light * 1e-3  # speed of light in km/s
        wave = other.wave.copy()
        for i in range(len(wave)):
            rvel = other.vrad[i]
            rv_factor = np.sqrt((1 - rvel / c_light) / (1 + rvel / c_light))
            wave[i] *= rv_factor
        wave = wave.ravel()

        line_mask = other.mask_line.ravel()
        cont_mask = other.mask_cont.ravel()

        for seg in range(self.nseg):
            # We simply interpolate between the masks, if most if the new pixel was
            # continuum / line mask then it will become that, otherwise bad
            rvel = self.vrad[seg]
            rv_factor = np.sqrt((1 - rvel / c_light) / (1 + rvel / c_light))
            w = self.wave[seg] * rv_factor
            cm = np.interp(w, wave, cont_mask) > 0.5
            lm = np.interp(w, wave, line_mask) > 0.5
            if keep_bpm:
                bpm = self.mask_bad[seg]
                cm[bpm] = False
                lm[bpm] = False
            self.mask[seg][cm] |= MASK_VALUES.CONT
            self.mask[seg][lm] |= MASK_VALUES.LINE
            self.mask[seg][~(cm | lm)] = MASK_VALUES.BAD
        return self

    def citation(self, output="string"):
        """Create a citation string for use in papers, or
        other places. The citations are from all components that
        contribute to the SME structure. SME and PySME, the linelist,
        the abundance, the atmosphere, and the NLTE grids.
        The default output is plaintext, but
        it is also possible to get bibtex format.

        Parameters
        ----------
        output : str, optional
            the output format, options are: ["string", "bibtex", "html", "markdown"], by default "string"

        Returns
        -------
        citation : str
            citation string in the desired output format
        """
        citation = [self.citation_info]
        citation += [self.atmo.citation_info]
        citation += [self.abund.citation_info]
        citation += [self.linelist.citation_info]
        citation += [self.nlte.citation_info]
        citation = "\n".join(citation)

        return self.create_citation(citation, output=output)

    def save(self, filename, format="flex", _async=False):
        """Save the whole SME structure to disk.

        The file format is zip file, with one info.json
        file for simple values, and numpy save files for
        large arrays. Substructures (linelist, abundance, etc.)
        have their own folder within the zip file.

        Parameters
        ----------
        filename : str
            filename to save the structure at
        compressed : bool, optional
            whether to compress the output, by default False
        """
        persistence.save(filename, self, format=format, _async=_async)

    @staticmethod
    def load(filename):
        """
        Load SME data from disk

        Currently supported file formats:
            * ".npy": Numpy save file of an SME_Struct
            * ".sav", ".inp", ".out": IDL save file with an sme structure
            * ".ech": Echelle file from (Py)REDUCE

        Parameters
        ----------
        filename : str, optional
            name of the file to load (default: 'sme.npy')

        Returns
        -------
        sme : SME_Struct
            Loaded SME structure

        Raises
        ------
        ValueError
            If the file format extension is not recognized
        """
        logger.info("Loading SME file %s", filename)
        ext = os.path.splitext(filename)[1]
        if ext == ".sme":
            s = SME_Structure()
            return persistence.load(filename, s)
        elif ext == ".npy":
            # Numpy Save file
            s = np.load(filename, allow_pickle=True)
            return np.atleast_1d(s)[0]
        elif ext == ".npz":
            s = np.load(filename, allow_pickle=True)
            return s["sme"][()]
        elif ext in [".sav", ".out", ".inp"]:
            # IDL save file (from SME)
            s = readsav(filename)["sme"]

            def unfold(obj):
                if isinstance(obj, bytes):
                    return obj.decode()
                elif isinstance(obj, np.recarray):
                    return {
                        name.casefold(): unfold(obj[name][0])
                        for name in obj.dtype.names
                    }
                return obj

            s = unfold(s)
            return SME_Structure(**s)
        elif ext == ".ech":
            # Echelle file (from REDUCE)
            ech = echelle.read(filename)
            s = SME_Structure()
            s.wave = [np.ma.compressed(w) for w in ech.wave]
            s.spec = [np.ma.compressed(s) for s in ech.spec]
            s.uncs = [np.ma.compressed(s) for s in ech.sig]

            for i, w in enumerate(s.wave):
                sort = np.argsort(w)
                s.wave[i] = w[sort]
                s.spec[i] = s.spec[i][sort]
                s.uncs[i] = s.uncs[i][sort]

            s.mask = [np.full(i.size, 1) for i in s.spec]
            s.mask[s.spec == 0] = MASK_VALUES.BAD
            s.wran = [[w[0], w[-1]] for w in s.wave]
            s.abund = Abund.solar()
            try:
                s.object = ech.head["OBJECT"]
            except KeyError:
                pass
            return s
        else:
            options = [".npy", ".sav", ".out", ".inp", ".ech"]
            raise ValueError(
                f"File format not recognised, expected one of {options} but got {ext}"
            )



# -*- coding: utf-8 -*-
import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.widgets import Button, SpanSelector
from scipy.constants import c

clight = c * 1e-3

# from ..sme.sme import SME_Struct

from .plot_colors import PlotColors

fmt = PlotColors()


class MaskPlot:
    """A plot that can be used to define the mask"""

    # Controls:
    # a, d keys: Switch between segments
    # Left, Right Mouse button: Select sections to change the mask depending on current mode
    #       mode == "good/bad" : left  -> line mask
    #                            right -> bad mask
    #                            Does not override existing good line mask
    #
    #       mode == "line/cont" : left  -> line mask
    #                             right -> continuum mask
    #                             Does not change bad line mask
    # Shift key : Switch between "good/bad" and "line/cont" modes

    def __init__(self, sme, segment=0, axes=None, show=True):
        self.wave = sme.wave
        self.spec = sme.spec
        self.mask = sme.mask
        self.smod = sme.synth
        self.segment = segment
        self.nsegments = len(self.wave)
        self.mode = "line/cont"
        self.lines = sme.linelist
        self.vrad = sme.vrad
        self.vrad = [v if v is not None else 0 for v in self.vrad]

        self.line_plot = None
        self.lock = False

        if axes is None:
            self.im = plt.subplots()[1]
        else:
            self.im = axes

        self.selector_line = SpanSelector(
            self.im,
            self.section_line_callback,
            direction="horizontal",
            useblit=True,
            button=(1,),
        )

        self.selector_cont = SpanSelector(
            self.im,
            self.section_continuum_callback,
            direction="horizontal",
            useblit=True,
            button=(3,),
        )

        self.im.figure.canvas.mpl_connect("key_press_event", self.key_event)
        self.im.callbacks.connect("xlim_changed", self.resize_event)

        ax_next = plt.axes([0.8, 0.025, 0.1, 0.04])
        self.button_next = Button(ax_next, "-->")
        self.button_next.on_clicked(self.next_segment)

        ax_prev = plt.axes([0.7, 0.025, 0.1, 0.04])
        self.button_prev = Button(ax_prev, "<--")
        self.button_prev.on_clicked(self.previous_segment)

        self.plot()
        if show:
            plt.show()

    def resize_event(self, event):
        if self.line_plot is not None and not self.lock:
            xlim = np.array(self.im.get_xlim())
            xlim *= 1 - self.vrad[self.segment] / clight
            idx = (self.lines_segment.wlcent >= xlim[0]) & (
                self.lines_segment.wlcent <= xlim[1]
            )
            importance = self.lines_segment.depth - min(self.lines_segment.depth[idx])
            if max(importance[idx]) != 0:
                importance /= max(importance[idx])
            else:
                importance[idx] = 1

            for i in np.where(idx)[0]:
                self.line_plot[i][0].set_visible(True)
                self.line_plot[i][1].set_visible(True)
                self.line_plot[i][0].set_alpha(importance[i])
                self.line_plot[i][1].set_alpha(importance[i])

            for i in np.where(~idx)[0]:
                self.line_plot[i][0].set_visible(False)
                self.line_plot[i][1].set_visible(False)

    def key_event(self, event):
        if event.key in ["shift"]:
            if self.mode == "good/bad":
                self.mode = "line/cont"
            else:
                self.mode = "good/bad"
            print("Switch to mode: %s" % self.mode)

        if event.key in ["a", "left"]:
            self.goto_segment(self.segment - 1)
        if event.key in ["d", "right"]:
            self.goto_segment(self.segment + 1)

    def section_line_callback(self, min, max):
        mask_type = "line" if self.mode == "line/cont" else "good"
        self.section_select(min, max, mask_type)

    def section_continuum_callback(self, min, max):
        mask_type = "cont" if self.mode == "line/cont" else "bad"
        self.section_select(min, max, mask_type)

    def section_select(self, min, max, mask_type):
        print("{} {:.3f} - {:.3f}".format(mask_type, min, max))
        # find points
        idx = (self.wave[self.segment] <= max) & (self.wave[self.segment] >= min)

        # update masks
        if mask_type == "line":
            mask_value = 1
        elif mask_type == "cont":
            mask_value = 2
        elif mask_type == "bad":
            mask_value = 0
        elif mask_type == "good":
            mask_value = 1

        if mask_type in ["line", "cont"]:
            idx = idx & (self.mask[self.segment] != 0)
        if mask_type == "good":
            idx = idx & (self.mask[self.segment] == 0)
        self.mask[self.segment][idx] = mask_value

        # update plot
        self.lock = True
        self.update()
        self.lock = False

    def plot(self, update=False):
        if self.mask is not None:
            mask = self.mask[self.segment]

        if self.spec is not None and not update:
            self.im.plot(
                self.wave[self.segment],
                self.spec[self.segment],
                label="Observation",
                **fmt["Obs"],
            )

        if self.smod is not None and not update:
            self.im.plot(
                self.wave[self.segment],
                self.smod[self.segment],
                label="Synthethic",
                **fmt["Syn"],
            )

        if self.spec is not None:
            self.fill_line = self.im.fill_between(
                self.wave[self.segment],
                0,
                self.spec[self.segment],
                where=mask == 1,
                label="Mask Line",
                **fmt["LineMask"],
            )

            m = mask == 2
            m[1:] = m[:-1] | m[1:]
            m[:-1] = m[:-1] | m[1:]
            self.fill_cont = self.im.fill_between(
                self.wave[self.segment],
                0,
                self.spec[self.segment],
                where=m,
                label="Mask Continuum",
                **fmt["ContMask"],
            )

        try:
            if self.lines is not None and not update:
                self.lock = True
                xlim = self.wave[self.segment][[0, -1]]
                xlim *= 1 - self.vrad[self.segment] / clight
                self.lines_segment = self.lines[
                    (self.lines.wlcent >= xlim[0]) & (self.lines.wlcent <= xlim[1])
                ]

                importance = self.lines_segment.depth - min(self.lines_segment.depth)
                importance /= max(importance)
                self.line_plot = [[None, None] for _ in self.lines_segment]
                for i, line in enumerate(self.lines_segment):
                    # if i > threshold:
                    wl = line.wlcent * (1 + self.vrad[self.segment] / clight)
                    self.line_plot[i][0] = self.im.text(
                        wl,
                        1.1,
                        f"{line.species} {line.wlcent:.2f}",
                        rotation="vertical",
                        horizontalalignment="right",
                        verticalalignment="top",
                        alpha=importance[i],
                    )
                    if self.spec is not None:
                        depth = np.interp(
                            wl,
                            self.wave[self.segment],
                            self.spec[self.segment],
                        )
                    else:
                        depth = np.interp(
                            wl,
                            self.wave[self.segment],
                            self.smod[self.segment],
                        )
                    self.line_plot[i][1] = self.im.vlines(
                        wl, ymin=depth, ymax=1.1, alpha=importance[i]
                    )
                self.lock = False
        except ValueError:
            pass

        self.im.figure.suptitle("SME Fit\nSegment %i" % self.segment)
        self.im.set_xlabel("Wavelength [Å]")
        self.im.set_ylabel("normalized Intensity")
        self.im.set_ylim((0, 1.2))
        self.im.set_xlim(self.im.get_xlim())
        self.im.legend(loc="lower left")

        self.im.figure.canvas.draw()

    def update(self, reset_view=False):
        if not reset_view:
            xlim = self.im.get_xlim()
            ylim = self.im.get_ylim()

        # Remove filled between
        if reset_view:
            self.im.collections.clear()
        elif isinstance(self.im.collections[0], mpl.collections.PolyCollection):
            del self.im.collections[:2]
        else:
            del self.im.collections[-2:]
        # del self.im.collections[:2]
        self.plot(update=True)

        if not reset_view:
            self.im.set_xlim(xlim)
            self.im.set_ylim(ylim)

    def connect_axes(self):
        self.im.callbacks.connect("xlim_changed", self.resize_event)

    def goto_segment(self, segment):
        if segment >= 0 and segment < len(self.wave) - 1:
            self.segment = segment
            self.im.cla()
            self.plot()
            self.connect_axes()

    def next_segment(self, _=None):
        self.goto_segment(self.segment + 1)

    def previous_segment(self, _=None):
        self.goto_segment(self.segment - 1)



# -*- coding: utf-8 -*-
from os.path import dirname

import numpy as np
import pytest

from pysme.sme import SME_Structure as SME_Struct
from pysme.solve import solve

cwd = dirname(__file__)
filename = "{}/testcase1.inp".format(cwd)


def test_simple():
    sme = SME_Struct.load(filename)
    sme2 = solve(sme, ["teff"])

    assert sme2.synth is not None
    assert sme2.fitresults is not None
    assert sme2.fitresults.covariance is not None
    assert isinstance(sme2.fitresults.covariance, np.ndarray)
    assert np.all(sme2.fitresults.covariance != 0)

    assert isinstance(sme2.fitresults.uncertainties, np.ndarray)
    assert len(sme2.fitresults.uncertainties) == 1
    assert sme2.fitresults.parameters[0] == "teff"
    assert sme2.fitresults.uncertainties[0] != 0

    assert np.array_equal(sme2.fitresults.covariance.shape, [1, 1])
    assert sme2.fitresults.covariance.ndim == 2

    assert sme2.fitresults.chisq is not None
    assert sme2.fitresults.chisq != 0
